#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generic,
    Hashable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

import torch
from torch import Tensor
from typing_extensions import NotRequired

from torchoutil.core.get import DeviceLike, get_device
from torchoutil.pyoutil.hashlib import HashName, hash_file
from torchoutil.pyoutil.logging import warn_once
from torchoutil.utils.saving.json import load_json, to_json
from torchoutil.utils.saving.load_fn import LOAD_FNS, LoadFnLike

T_Hashable = TypeVar("T_Hashable", bound=Hashable)

pylog = logging.getLogger(__name__)


class RegistryEntry(TypedDict):
    url: str
    fname: str
    hash_value: NotRequired[str]
    hash_type: NotRequired[HashName]
    state_dict_key: NotRequired[str]
    architecture: NotRequired[str]


class RegistryHub(Generic[T_Hashable]):
    def __init__(
        self,
        infos: Mapping[T_Hashable, RegistryEntry],
        register_root: Union[str, Path, None] = None,
    ) -> None:
        """
        Args:
            infos: Maps model_name to their checkpoint information, with download url, filename, hash value, hash type and state_dict key.
            register_root: Directory where checkpoints are saved. If None, defaults to `~/.cache/torch/hub/checkpoints`.
        """
        infos = dict(infos.items())
        if register_root is None:
            register_root = get_default_register_root()
        else:
            register_root = Path(register_root)

        super().__init__()
        self._infos = infos
        self._ckpt_parent_path = register_root

    @property
    def infos(self) -> Dict[T_Hashable, RegistryEntry]:
        return self._infos

    @property
    def register_root(self) -> Path:
        return self._ckpt_parent_path.resolve()

    @property
    def names(self) -> List[T_Hashable]:
        return list(self._infos.keys())

    @property
    def paths(self) -> List[Path]:
        return [self.get_path(model_name) for model_name in self.names]

    def get_path(self, name: T_Hashable) -> Path:
        if name not in self.names:
            msg = f"Invalid argument {name=}. (expected one of {self.names})"
            raise ValueError(msg)

        fname = self._infos[name]["fname"]
        fpath = self.register_root.joinpath(fname)
        return fpath

    def load_state_dict(
        self,
        name_or_path: Union[T_Hashable, str, Path],
        *,
        device: DeviceLike = None,
        offline: bool = False,
        load_fn: LoadFnLike = torch.load,
        load_kwds: Optional[Dict[str, Any]] = None,
        verbose: int = 0,
    ) -> Dict[str, Tensor]:
        """Load state_dict weights.

        Args:
            model_name_or_path: Model name (case sensitive) or path to checkpoint file.
            device: Device of checkpoint weights. (deprecated)
            offline: If False, the checkpoint from a model name will be automatically downloaded.
            load_fn: Load function backend. defaults to torch.load.
            load_kwds: Optional keywords arguments passed to load_fn. defaults to None.
            verbose: Verbose level. defaults to 0.

        Returns:
            Loaded file content.
        """
        if isinstance(load_fn, str):
            if load_fn not in LOAD_FNS:
                msg = f"Invalid argument {load_fn=}. (expected one of {tuple(LOAD_FNS.keys())})"
                raise ValueError(msg)
            load_fn = LOAD_FNS[load_fn]

        if load_kwds is None:
            load_kwds = {}

        if device is not None:
            src_device = device
            device = get_device(device)
            msg = f"Deprecated argument device={src_device}. Use `load_kwds=dict(map_location={device})` with function torch.load instead."
            warn_once(msg, __name__)

            if device is not None:
                load_kwds["map_location"] = device

        if isinstance(name_or_path, (str, Path)) and osp.isfile(name_or_path):
            path = Path(name_or_path)
            name = self._get_name(path)
        else:
            name = name_or_path
            try:
                path = self.get_path(name_or_path)  # type: ignore
            except ValueError:
                msg = f"Invalid argument {name_or_path=}. (expected a path to a checkpoint file or a model name in {self.names})"
                raise ValueError(msg)

            if path.is_file():
                pass
            elif offline:
                msg = f"Cannot find checkpoint model file in '{path}' for model '{name_or_path}' with mode {offline=}."
                raise FileNotFoundError(msg)
            else:
                self.download_file(name_or_path, verbose=verbose)  # type: ignore

        del name_or_path

        info = self._infos.get(name, {})  # type: ignore
        state_dict_key = info.get("state_dict_key", None)
        data = load_fn(path, **load_kwds)

        if state_dict_key is None:
            result = data
        else:
            result = data[state_dict_key]

        if verbose >= 1:
            test_map = data.get("test_mAP", "unknown")
            msg = f"Loading encoder weights from '{path}'... (with test_mAP={test_map})"
            pylog.info(msg)

        return result

    def download_file(
        self,
        name: T_Hashable,
        force: bool = False,
        check_hash: bool = True,
        verbose: int = 0,
    ) -> Tuple[Path, bool]:
        """Download checkpoint file."""
        model_path = self.get_path(name)
        exists = model_path.exists()

        if exists and not force:
            return model_path, False

        if exists and force:
            os.remove(model_path)

        model_path.parent.mkdir(parents=True, exist_ok=True)
        url = self._infos[name]["url"]
        torch.hub.download_url_to_file(url, str(model_path), progress=verbose >= 1)

        if not check_hash:
            return model_path, True

        valid = self.is_valid_hash(name)

        if valid:
            return model_path, True
        else:
            raise ValueError(f"Invalid hash for file '{model_path}'.")

    def is_valid_hash(
        self,
        name: T_Hashable,
    ) -> bool:
        info = self.infos[name]
        if "hash_type" not in info or "hash_value" not in info:
            msg = f"Cannot check hash for {name}. (cannot find any expected hash value or type)"
            pylog.warning(msg)
            return True

        hash_type = info["hash_type"]
        expected_hash_value = info["hash_value"]

        model_path = self.get_path(name)
        hash_value = hash_file(model_path, hash_type)
        valid = hash_value == expected_hash_value
        return valid

    def save(self, path: Union[str, Path]) -> None:
        """Save info to JSON file."""
        args = {
            "infos": self._infos,
            "register_root": str(self._ckpt_parent_path),
        }
        to_json(args, path)

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> "RegistryHub":
        """Load register info from JSON file."""
        args = load_json(path)
        return RegistryHub(**args)

    def _get_name(self, path: Union[str, Path]) -> Optional[T_Hashable]:
        path_to_name = {
            path_i.resolve().expanduser(): name_i
            for path_i, name_i in zip(self.paths, self.names)
        }
        path = Path(path).resolve().expanduser()
        name = path_to_name.get(path, None)
        return name


def get_default_register_root() -> Path:
    """Default register root path is `~/.cache/torch/hub/checkpoints`, which is based on `torch.hub.get_dir`."""
    path = torch.hub.get_dir()
    path = Path(path)
    path = path.joinpath("checkpoints")
    return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib
import inspect
import logging
import pkgutil
import sys
from importlib.util import find_spec
from types import ModuleType
from typing import Iterable, List

DEFAULT_SKIPPED = (
    "reimport_all",
    "get_ipython",
    "exit",
    "quit",
    "__name__",
    "__doc__",
    "__package__",
    "__loader__",
    "__spec__",
    "__builtin__",
    "__builtins__",
)

pylog = logging.getLogger(__name__)


def package_is_available(package: str) -> bool:
    """Returns True if package is installed in the current python environment."""
    try:
        return find_spec(package) is not None
    except AttributeError:
        # Old support for Python <= 3.6
        return False
    except (ImportError, ModuleNotFoundError):
        # Python >= 3.7
        return False


def search_imported_submodules(
    root: ModuleType,
    parent_name: str = "",
) -> List[ModuleType]:
    """Return the submodules already imported."""
    if root.__package__ is None:
        raise ValueError(f"Cannot search in module '{root}'. (found __package__=None)")

    if parent_name != "":
        parent_name = ".".join([parent_name, root.__package__])
    else:
        parent_name = root.__package__

    if hasattr(root, "__path__"):
        paths = root.__path__
    elif root.__file__ is not None:
        paths = [root.__file__]
    else:
        raise ValueError

    module_infos = list(pkgutil.iter_modules(paths))
    candidates = []

    for info in module_infos:
        if info.name == "__main__":
            continue
        fullname = f"{parent_name}.{info.name}"
        candidate = sys.modules.get(fullname)
        if candidate is None:
            continue
        sub_candidates = search_imported_submodules(candidate, parent_name)
        candidates += sub_candidates + [candidate]

    return candidates


def reload_submodules(root: ModuleType, verbose: int = 0) -> List[ModuleType]:
    candidates = search_imported_submodules(root) + [root]
    for candidate in candidates:
        if verbose > 0:
            pylog.info(f"Reload '{candidate}'")
        importlib.reload(candidate)
    return candidates


def reload_globals_modules(
    skipped: Iterable[str] = DEFAULT_SKIPPED,
    verbose: int = 0,
) -> List[ModuleType]:
    """Re-import modules and functions in the caller context. This function does not work with builtins constants values."""
    skipped = dict.fromkeys(skipped)

    importlib.invalidate_caches()
    caller_globals = dict(inspect.getmembers(inspect.stack()[1][0]))["f_globals"]
    if verbose >= 1:
        print(f"{caller_globals.keys()=}")

    candidates = []

    for k, v in caller_globals.items():
        if k in skipped:
            continue

        if isinstance(v, ModuleType):
            candidates += reload_submodules(v, verbose=verbose)
            continue

        v = inspect.getmodule(v)
        if v is None or v.__name__ == "__main__":
            continue

        candidates += reload_submodules(v, verbose=verbose)

        try:
            caller_globals[k] = getattr(v, k)
        except AttributeError:
            if verbose >= 1:
                print(f"Cannot set parent global value '{k}'.")

    return candidates



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    TypeVar,
    Union,
    overload,
)

import torch
from torch import Generator, Tensor

from torchoutil.core.get import (
    DeviceLike,
    DTypeLike,
    get_device,
    get_dtype,
    get_generator,
)
from torchoutil.extras.numpy import np
from torchoutil.nn.functional.crop import crop_dim
from torchoutil.nn.functional.pad import PadMode, PadValue, pad_dim
from torchoutil.pyoutil.collections import all_eq
from torchoutil.pyoutil.collections import flatten as builtin_flatten
from torchoutil.pyoutil.collections import prod as builtin_prod
from torchoutil.pyoutil.functools import identity  # noqa: F401
from torchoutil.pyoutil.typing import T_BuiltinScalar
from torchoutil.types import is_number_like, is_scalar_like
from torchoutil.types._typing import (
    BuiltinNumber,
    NumberLike,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
)

T = TypeVar("T")
U = TypeVar("U")

PadCropAlign = Literal["left", "right", "center", "random"]
PAD_CROP_ALIGN_VALUES = ("left", "right", "center", "random")


def repeat_interleave_nd(x: Tensor, repeats: int, dim: int = 0) -> Tensor:
    """Generalized version of torch.repeat_interleave for N >= 1 dimensions.
    The output size will be (..., D*repeats, ...), where D is the size of the dimension of the dim argument.

    Args:
        x: Any tensor of shape (..., D, ...) with at least 1 dim.
        repeats: Number of repeats.
        dim: The dimension to repeat. defaults to 0.

    Examples::
    ----------
        >>> x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        >>> repeat_interleave_nd(x, n=2, dim=0)
        tensor([[0, 1, 2, 3],
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [4, 5, 6, 7]])
    """
    if x.ndim == 0:
        msg = f"Function repeat_interleave_nd does not supports 0-d tensors. (found {x.ndim=} == 0)"
        raise ValueError(msg)

    dim = dim % x.ndim
    x = x.unsqueeze(dim=dim + 1)
    shape = list(x.shape)
    shape[dim + 1] = repeats
    x = x.expand(*shape)
    x = x.flatten(dim, dim + 1)
    return x


def resample_nearest_rates(
    x: Tensor,
    rates: Union[float, Iterable[float]],
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    """Nearest neigbour resampling using tensor slices.

    Args:
        x: Input tensor.
        rate: The reduction factor of each axis, e.g. a factor of 0.5 will divide the input axes by 2.
        dims: Dimensions to apply resampling. defaults to -1.
        round_fn: Rounding function to compute sub-indices. defaults to torch.floor.
    """
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    if isinstance(rates, (int, float)):
        steps = [1.0 / rates] * len(dims)
    else:
        steps = [1.0 / rate for rate in rates]

    return resample_nearest_steps(
        x,
        steps,
        dims=dims,
        round_fn=round_fn,
    )


def resample_nearest_freqs(
    x: Tensor,
    orig_freq: int,
    new_freq: int,
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    return resample_nearest_steps(
        x,
        orig_freq / new_freq,
        dims=dims,
        round_fn=round_fn,
    )


def resample_nearest_steps(
    x: Tensor,
    steps: Union[float, Iterable[float]],
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    if isinstance(steps, (int, float)):
        steps = [steps] * len(dims)
    else:
        steps = list(steps)  # type: ignore
        if len(steps) != len(dims):
            raise ValueError(f"Invalid arguments sizes {len(steps)=} != {len(dims)}.")

    slices: List[Union[slice, Tensor]] = [slice(None)] * x.ndim

    for dim, step in zip(dims, steps):
        length = x.shape[dim]
        indexes = torch.arange(0, length, step)
        indexes = round_fn(indexes).long().clamp(min=0, max=length - 1)
        slices[dim] = indexes

    x = x[slices]
    return x


def transform_drop(
    transform: Callable[[T], T],
    x: T,
    p: float,
    generator: Union[int, Generator, None] = None,
) -> T:
    """Apply a transform on a tensor with a probability of p.

    Args:
        transform: Transform to apply.
        x: Argument of the transform.
        p: Probability p to apply the transform. Cannot be negative.
            If > 1, it will apply the transform floor(p) times and apply a last time with a probability of p - floor(p).
    """
    if p < 0.0:
        raise ValueError(f"Invalid argument {p=} < 0")
    generator = get_generator(generator)

    p_floor = math.floor(p)
    for _ in range(p_floor):
        x = transform(x)

    sampled = torch.rand((), generator=generator)
    if sampled + p_floor < p:
        x = transform(x)

    return x


def pad_and_crop_dim(
    x: Tensor,
    target_length: int,
    *,
    align: PadCropAlign = "left",
    pad_value: PadValue = 0.0,
    dim: int = -1,
    mode: PadMode = "constant",
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Pad and crop along the specified dimension."""
    x = pad_dim(
        x,
        target_length,
        align=align,
        pad_value=pad_value,
        dim=dim,
        mode=mode,
        generator=generator,
    )
    x = crop_dim(
        x,
        target_length,
        align=align,
        dim=dim,
        generator=generator,
    )
    return x


def shuffled(
    x: Tensor,
    dims: Union[int, Iterable[int]] = -1,
    generator: Union[int, Generator, None] = None,
) -> Tensor:
    """Returns a shuffled version of the input tensor along specific dimension(s)."""
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    generator = get_generator(generator)
    slices: List[Union[slice, Tensor]] = [slice(None) for _ in range(x.ndim)]
    for dim in dims:
        indices = torch.randperm(x.shape[dim], generator=generator)
        slices[dim] = indices
    x = x[slices]
    return x


@overload
def flatten(
    x: Tensor,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> Tensor1D:
    ...


@overload
def flatten(
    x: Union[np.ndarray, np.generic],
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> np.ndarray:
    ...


@overload
def flatten(
    x: T_BuiltinScalar,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[T_BuiltinScalar]:
    ...


@overload
def flatten(
    x: Iterable[T_BuiltinScalar],
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[T_BuiltinScalar]:
    ...


@overload
def flatten(
    x: Any,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[Any]:
    ...


def flatten(
    x: Any,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> Any:
    if isinstance(x, Tensor):
        end_dim = end_dim if end_dim is not None else x.ndim - 1
        return x.flatten(start_dim, end_dim)
    elif isinstance(x, np.generic):
        return x.flatten()
    elif isinstance(x, np.ndarray):
        if start_dim == 0 and (end_dim is None or end_dim >= x.ndim - 1):
            return x.flatten()
        else:
            end_dim = end_dim if end_dim is not None else x.ndim - 1
            shape = list(x.shape)
            shape = (
                shape[:start_dim]
                + [builtin_prod(shape[start_dim : end_dim + 1])]
                + shape[end_dim + 1 :]
            )
            return x.reshape(*shape)
    else:
        return builtin_flatten(x, start_dim, end_dim, is_scalar_fn=is_scalar_like)


@overload
def to_tensor(
    data: BuiltinNumber,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor0D:
    ...


@overload
def to_tensor(
    data: Sequence[BuiltinNumber],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor1D:
    ...


@overload
def to_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor2D:
    ...


@overload
def to_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor3D:
    ...


@overload
def to_tensor(data: Any, dtype: DTypeLike = None, device: DeviceLike = None) -> Tensor:
    ...


def to_tensor(data: Any, dtype: DTypeLike = None, device: DeviceLike = None) -> Tensor:
    """Convert arbitrary data to tensor. Unlike `torch.as_tensor`, it works recursively and stack sequences like List[Tensor].

    Args:
        data: Data to convert to tensor. Can be Tensor, np.ndarray, list, tuple or any number-like object.
        dtype: Target torch dtype.
        device: Target torch device.

    Returns:
        PyTorch tensor created from data.
    """
    if isinstance(data, (Tensor, np.ndarray)) or is_number_like(data):
        dtype = get_dtype(dtype)
        device = get_device(device)
        return torch.as_tensor(data, dtype=dtype, device=device)

    elif isinstance(data, (list, tuple)):
        tensors = [to_tensor(data_i, dtype=dtype, device=device) for data_i in data]
        shapes = [tensor.shape for tensor in tensors]
        if not all_eq(shapes):
            msg = f"Cannot convert to tensor a list of elements with heterogeneous shapes. (found {shapes})"
            raise ValueError(msg)
        return torch.stack(tensors)

    else:
        EXPECTED = (Tensor, np.ndarray, NumberLike, list, tuple)
        msg = f"Invalid argument type '{type(data)}'. (expected one of {EXPECTED})"
        raise TypeError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Module used for typing when numpy is not installed."""

from typing import Any

from torchoutil.pyoutil.inspect import get_current_fn_name


class _placeholder:
    def __init__(self, *args, **kwargs) -> None:
        ...

    def __getattr__(self, name: str) -> Any:
        return self


class generic(_placeholder):
    ...


class number(generic):
    ...


class bool_(generic):
    ...


class dtype(_placeholder):
    ...


class ndarray(_placeholder):
    def __getitem__(self, *args) -> Any:
        return self


def array(x: Any, *args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependancy 'numpy' is not installed. Please install it using 'pip install torchoutil[extras]'"
    raise NotImplementedError(msg)


def asarray(x: Any, *args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependancy 'numpy' is not installed. Please install it using 'pip install torchoutil[extras]'"
    raise NotImplementedError(msg)


def iscomplexobj(x: Any):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependancy 'numpy' is not installed. Please install it using 'pip install torchoutil[extras]'"
    raise NotImplementedError(msg)


def empty(*args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependancy 'numpy' is not installed. Please install it using 'pip install torchoutil[extras]'"
    raise NotImplementedError(msg)


class complex64(dtype):
    ...


class complex128(dtype):
    ...


class complex256(dtype):
    ...


class float16(dtype):
    ...


class float32(dtype):
    ...


class float64(dtype):
    ...


class float128(dtype):
    ...


class floating(dtype):
    ...


class int16(dtype):
    ...


class int32(dtype):
    ...


class int64(dtype):
    ...


class uint8(dtype):
    ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Generic, List, Mapping, Optional, Sequence, Union

import torch
from torch import Tensor, nn

from torchoutil.core.get import DeviceLike, DTypeLike
from torchoutil.nn.functional.multiclass import (
    T_Name,
    index_to_name,
    index_to_onehot,
    name_to_index,
    name_to_onehot,
    onehot_to_index,
    onehot_to_name,
    probs_to_index,
    probs_to_name,
    probs_to_onehot,
)
from torchoutil.pyoutil.collections import dump_dict


class IndexToOnehot(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.index_to_onehot`.
    """

    def __init__(
        self,
        num_classes: int,
        *,
        padding_idx: Optional[int] = None,
        device: DeviceLike = None,
        dtype: DTypeLike = torch.bool,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.padding_idx = padding_idx
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        index: Union[List[int], Tensor],
    ) -> Tensor:
        onehot = index_to_onehot(
            index,
            self.num_classes,
            padding_idx=self.padding_idx,
            device=self.device,
            dtype=self.dtype,
        )
        return onehot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                num_classes=self.num_classes,
                padding_idx=self.padding_idx,
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class IndexToName(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.index_to_name`.
    """

    def __init__(
        self,
        idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        index: Union[List[int], Tensor],
    ) -> List[T_Name]:
        name = index_to_name(index, self.idx_to_name)
        return name


class OnehotToIndex(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.onehot_to_index`.
    """

    def __init__(self, dim: int = -1) -> None:
        super().__init__()
        self.dim = dim

    def forward(
        self,
        onehot: Tensor,
    ) -> Tensor:
        index = onehot_to_index(onehot, dim=self.dim)
        return index

    def extra_repr(self) -> str:
        return dump_dict(dict(dim=self.dim))


class OnehotToName(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.onehot_to_name`.
    """

    def __init__(
        self,
        idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
        dim: int = -1,
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name
        self.dim = dim

    def forward(
        self,
        onehot: Tensor,
    ) -> List[T_Name]:
        name = onehot_to_name(onehot, self.idx_to_name, dim=self.dim)
        return name

    def extra_repr(self) -> str:
        return dump_dict(dict(dim=self.dim))


class NameToIndex(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.name_to_index`.
    """

    def __init__(
        self,
        idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        name: List[T_Name],
    ) -> Tensor:
        index = name_to_index(name, self.idx_to_name)
        return index


class NameToOnehot(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.name_to_onehot`.
    """

    def __init__(
        self,
        idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
        *,
        device: DeviceLike = None,
        dtype: DTypeLike = torch.bool,
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        name: List[T_Name],
    ) -> Tensor:
        onehot = name_to_onehot(
            name,
            self.idx_to_name,
            device=self.device,
            dtype=self.dtype,
        )
        return onehot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class ProbsToIndex(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.probs_to_index`.
    """

    def __init__(self, dim: int = -1) -> None:
        super().__init__()
        self.dim = dim

    def forward(
        self,
        probs: Tensor,
    ) -> Tensor:
        index = probs_to_index(probs, dim=self.dim)
        return index

    def extra_repr(self) -> str:
        return dump_dict(dict(dim=self.dim))


class ProbsToOnehot(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.probs_to_onehot`.
    """

    def __init__(
        self,
        *,
        dim: int = -1,
        device: DeviceLike = None,
        dtype: DTypeLike = torch.bool,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        probs: Tensor,
    ) -> Tensor:
        onehot = probs_to_onehot(
            probs,
            dim=self.dim,
            device=self.device,
            dtype=self.dtype,
        )
        return onehot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class ProbsToName(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multiclass.probs_to_name`.
    """

    def __init__(
        self,
        idx_to_name: Union[Mapping[int, T_Name], Sequence[T_Name]],
        dim: int = -1,
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name
        self.dim = dim

    def forward(
        self,
        probs: Tensor,
    ) -> List[T_Name]:
        name = probs_to_name(probs, self.idx_to_name, dim=self.dim)
        return name

    def extra_repr(self) -> str:
        return dump_dict(dict(dim=self.dim))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Callable, Dict, Literal, TypeVar, Union

import torch

from torchoutil.core.packaging import (
    _NUMPY_AVAILABLE,
    _SAFETENSORS_AVAILABLE,
    _YAML_AVAILABLE,
)
from torchoutil.extras.numpy import np
from torchoutil.utils.saving.csv import load_csv
from torchoutil.utils.saving.json import load_json
from torchoutil.utils.saving.pickle import load_pickle

T = TypeVar("T", covariant=True)


LoadFn = Callable[[Path], T]
LoadFnName = Literal["csv", "json", "numpy", "pickle", "safetensors", "torch", "yaml"]
LoadFnLike = Union[LoadFn[T], LoadFnName]


LOAD_FNS: Dict[LoadFnName, LoadFn[Any]] = {
    "csv": load_csv,
    "json": load_json,
    "pickle": load_pickle,
    "torch": torch.load,
}

if _NUMPY_AVAILABLE:
    LOAD_FNS["numpy"] = np.load


if _SAFETENSORS_AVAILABLE:
    from torchoutil.utils.saving.safetensors import load_safetensors

    LOAD_FNS["safetensors"] = load_safetensors


if _YAML_AVAILABLE:
    from torchoutil.utils.saving.yaml import load_yaml

    LOAD_FNS["yaml"] = load_yaml



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    ClassVar,
    Dict,
    Iterator,
    Protocol,
    Sized,
    Tuple,
    TypeVar,
    Union,
    runtime_checkable,
)

NoneType = type(None)
EllipsisType = type(...)

BuiltinCollection = Union[list, tuple, dict, set, frozenset]
BuiltinNumber = Union[bool, int, float, complex]
BuiltinScalar = Union[bool, int, float, complex, NoneType, str, bytes]

T = TypeVar("T", covariant=True)
T_BuiltinNumber = TypeVar("T_BuiltinNumber", bound=BuiltinNumber)
T_BuiltinScalar = TypeVar("T_BuiltinScalar", bound=BuiltinScalar)


@runtime_checkable
class DataclassInstance(Protocol):
    # Class meant for typing purpose only
    __dataclass_fields__: ClassVar[Dict[str, Any]]


@runtime_checkable
class NamedTupleInstance(Protocol):
    # Class meant for typing purpose only
    _fields: Tuple[str, ...]
    _field_defaults: Dict[str, Any]

    def _asdict(self) -> Dict[str, Any]:
        ...

    def __getitem__(self, idx, /):
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class SizedIterable(Protocol[T]):
    def __iter__(self) -> T:
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class SupportsLenAndGetItem(Protocol[T]):
    def __getitem__(self, idx, /) -> T:
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class SupportsLenAndGetItemAndIter(Protocol[T]):
    def __getitem__(self, idx, /) -> T:
        ...

    def __iter__(self) -> Iterator[T]:
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class SupportsBool(Protocol):
    def __bool__(self) -> bool:
        ...


BoolLike = Union[bool, int, SupportsBool, Sized]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import re
from functools import partial
from re import Pattern
from typing import Any, Callable, Iterable, List, Optional, TypeVar, Union

from torchoutil.pyoutil.collections import find

T = TypeVar("T")

PatternLike = Union[str, Pattern]
PatternListLike = Union[PatternLike, Iterable[PatternLike]]

pylog = logging.getLogger(__name__)


def compile_patterns(patterns: PatternListLike) -> List[Pattern]:
    """Compile patterns-like to a list."""
    if isinstance(patterns, (str, Pattern)):
        patterns = [patterns]
    patterns = [re.compile(pattern) for pattern in patterns]
    return patterns


def find_patterns(
    x: str,
    patterns: PatternListLike,
    *,
    match_fn: Callable[[PatternLike, str], Any] = re.search,
    default: T = -1,
) -> Union[int, T]:
    """Find index of a pattern that match the first argument. If no pattern matches, returns the default value (-1)."""
    patterns = compile_patterns(patterns)
    index = find(x, patterns, match_fn=match_fn, order="right", default=default)
    return index


def match_patterns(
    x: str,
    include: Optional[PatternListLike] = ".*",
    *,
    exclude: Optional[PatternListLike] = (),
    match_fn: Callable[[PatternLike, str], Any] = re.search,
) -> bool:
    """Returns True if the first argument match at least 1 included pattern and does not match any excluded pattern.

    Args:
        x: String to check.
        include: Acceptable pattern(s) for x. If None, match all patterns with '.*'. defaults to '.*'.
        exclude Forbidden pattern(s) for x. If None, match no patterns with value (). defaults to ().
        match_fn: Match function use to compare a pattern with argument x. defaults to re.search.
    """
    if include is None:
        include = ".*"
    if exclude is None:
        exclude = ()
    include_index = find_patterns(x, include, match_fn=match_fn, default=-1)
    exclude_index = find_patterns(x, exclude, match_fn=match_fn, default=-1)
    return include_index != -1 and exclude_index == -1


def get_key_fn(
    patterns: PatternListLike,
    *,
    match_fn: Callable[[PatternLike, str], Any] = re.search,
) -> Callable[[str], int]:
    """
    Usage:
    ```
    >>> lst = ["a", "abc", "aa", "abcd"]
    >>> patterns = ["^ab"]  # sort list with elements starting with 'ab' first
    >>> list(sorted(lst, key=get_key_fn(patterns)))
    ... ["abc", "abcd", "a", "aa"]
    ```
    """
    patterns = compile_patterns(patterns)
    key_fn = partial(
        find_patterns,
        patterns=patterns,
        match_fn=match_fn,
        default=len(patterns),
    )
    return key_fn  # type: ignore


def sort_with_patterns(
    x: Iterable[str],
    patterns: PatternListLike,
    *,
    match_fn: Callable[[PatternLike, str], Any] = re.search,
    reverse: bool = False,
) -> List[str]:
    key_fn = get_key_fn(patterns, match_fn=match_fn)
    x = sorted(x, key=key_fn, reverse=reverse)
    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Type used for backward compatibility."""

import sys
from typing import Generic, NamedTuple, Sequence

from torch import LongTensor, Tensor, __version__
from torch.torch_version import TorchVersion
from typing_extensions import TypeVar

T = TypeVar("T")


if __version__ < TorchVersion("2.0.0"):

    class namedtuple_values_indices(NamedTuple):
        values: Tensor
        indices: LongTensor

        def __init__(self, args: Sequence[Tensor]):
            super().__init__(*args)

    class min(namedtuple_values_indices):
        ...

    class max(namedtuple_values_indices):
        ...

    class topk(namedtuple_values_indices):
        ...

else:
    from torch.return_types import max, min, topk  # type: ignore # noqa: F401


if sys.version_info.major == 3 and sys.version_info.minor <= 8:
    # workaround for typing in python 3.8
    class _shape_base(NamedTuple):
        valid: bool
        shape: T  # type: ignore

    class shape(_shape_base, Generic[T]):  # type: ignore
        ...

else:

    class shape(NamedTuple, Generic[T]):
        valid: bool
        shape: T


class ndim(NamedTuple):
    valid: bool
    ndim: int



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import inspect
from typing import Any, get_args


def get_current_fn_name() -> str:
    try:
        return inspect.currentframe().f_back.f_code.co_name  # type: ignore
    except AttributeError:
        return ""


def get_fullname(x: Any, *, inst_suffix: str = "(...)") -> str:
    """Returns the classname of an object with parent modules.

    Args:
        obj: Object to scan.
        inst_suffix: Suffix appended to the classname in case the object is an instance of a class.

    Example 1
    ----------
    >>> get_fullname([0, 1, 2])
    ... 'builtins.list(...)'
    >>> get_fullname(1.0)
    ... 'builtins.float(...)'
    >>> class A: def f(self): return 0
    >>> a = A()
    >>> get_fullname(a)
    ... '__main__.A(...)'
    >>> get_fullname(A)
    ... '__main__.A'
    >>> get_fullname(a.f)
    ... '__main__.A.f'
    >>> get_fullname(A.f)
    ... '__main__.A.f'
    """
    if hasattr(x, "__module__") and hasattr(x, "__qualname__"):
        name = f"{x.__module__}.{x.__qualname__}"
    elif inspect.ismodule(x):
        name = x.__name__
    else:
        cls = x.__class__
        name = f"{cls.__module__}.{cls.__qualname__}{inst_suffix}"

    cls_args = get_args(x)
    if len(cls_args) != 0:
        argsnames = [get_fullname(arg, inst_suffix=inst_suffix) for arg in cls_args]
        argsnames_str = ", ".join(argsnames)
        name = f"{name}[{argsnames_str}]"

    return name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from typing import Any, Dict, List, Optional, TypeVar

import torch

from torchoutil.nn.functional.others import can_be_converted_to_tensor, can_be_stacked
from torchoutil.nn.functional.pad import pad_and_stack_rec
from torchoutil.pyoutil.collections import KeyMode, list_dict_to_dict_list
from torchoutil.pyoutil.re import PatternListLike, match_patterns

K = TypeVar("K")
V = TypeVar("V")


class CollateDict:
    """Collate class to handle a batch dict."""

    def __init__(self, key_mode: KeyMode = "same") -> None:
        super().__init__()
        self.key_mode: KeyMode = key_mode

    def __call__(self, batch_lst: List[Dict[K, V]]) -> Dict[K, List[V]]:
        result = list_dict_to_dict_list(
            batch_lst,
            key_mode=self.key_mode,
        )
        return result  # type: ignore


class AdvancedCollateDict:
    """Collate class to automatically convert to tensor, pad sequences and filter keys in a batch dict."""

    def __init__(
        self,
        pad_values: Optional[Dict[str, Any]] = None,
        include_keys: Optional[PatternListLike] = None,
        exclude_keys: Optional[PatternListLike] = None,
        key_mode: KeyMode = "same",
    ) -> None:
        """Collate list of dict into a dict of list WITH auto-padding for given keys."""
        if pad_values is None:
            pad_values = {}

        super().__init__()
        self.pad_values = pad_values
        self.include_keys = include_keys
        self.exclude_keys = exclude_keys
        self.key_mode: KeyMode = key_mode

    def __call__(self, batch_lst: List[Dict[str, Any]]) -> Dict[str, Any]:
        batch_dict = list_dict_to_dict_list(
            batch_lst,
            key_mode=self.key_mode,
        )
        batch_keys = [
            k
            for k in batch_dict.keys()
            if match_patterns(
                k,
                self.include_keys,
                exclude=self.exclude_keys,
                match_fn=re.match,
            )
        ]
        batch_dict = {k: batch_dict[k] for k in batch_keys}
        result = {}

        for key, values in batch_dict.items():
            if key in self.pad_values:
                values = pad_and_stack_rec(values, self.pad_values[key])
            elif can_be_stacked(values):
                values = torch.stack(values)
            elif can_be_converted_to_tensor(values):
                values = torch.as_tensor(values)

            result[key] = values

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from torch import Tensor


def softmax_multidim(
    x: Tensor,
    *,
    dims: Union[Iterable[int], None] = (-1,),
) -> Tensor:
    """A multi-dimensional version of torch.softmax along multiple dimensions at the same time."""
    x = x.exp()
    return log_softmax_multidim(x, dims=dims)


def log_softmax_multidim(
    x: Tensor,
    *,
    dims: Union[Iterable[int], None] = (-1,),
) -> Tensor:
    """A multi-dimensional version of torch.log_softmax along multiple dimensions at the same time."""
    if dims is not None:
        dims = tuple(dims)
    result = x / x.sum(dim=dims, keepdim=True)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import logging
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    MutableMapping,
    Optional,
    OrderedDict,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    overload,
    runtime_checkable,
)

import torch
import torch.utils
from torch import Tensor, nn
from torch.nn.parameter import Parameter
from typing_extensions import Concatenate, ParamSpec

from torchoutil.nn.functional.checksum import checksum_module
from torchoutil.nn.functional.others import count_parameters
from torchoutil.pyoutil.collections import dump_dict
from torchoutil.pyoutil.re import match_patterns
from torchoutil.pyoutil.typing import NoneType, is_dict_str, is_mapping_str

T = TypeVar("T", covariant=True)
InType = TypeVar("InType", covariant=False, contravariant=True)
OutType = TypeVar("OutType", covariant=True, contravariant=False)
OutType2 = TypeVar("OutType2", covariant=True, contravariant=False)
OutType3 = TypeVar("OutType3", covariant=False, contravariant=False)
T_MutableMappingStr = TypeVar("T_MutableMappingStr", bound=MutableMapping[str, Any])
P = ParamSpec("P")

DeviceDetectMode = Literal["proxy", "first_param", "none"]
DEVICE_DETECT_MODES = ("proxy", "first_param", "none")
_DEFAULT_DEVICE_DETECT_MODE = "first_param"


pylog = logging.getLogger(__name__)


__all__ = [
    "DeviceDetectMode",
    "DEVICE_DETECT_MODES",
    "EModule",
    "ESequential",
    "EModuleList",
    "EModuleDict",
    "EModulePartial",
]


@runtime_checkable
class SupportsTypedForward(Protocol[InType, OutType]):
    def __call__(self, *args, **kwargs):
        ...

    def forward(self, x: InType, /) -> OutType:
        ...


TypedModuleLike = Union[
    SupportsTypedForward[InType, OutType],
    "TypedModule[InType, OutType]",
]


class ProxyDeviceModule(nn.Module):
    def __init__(
        self,
        *,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        if device_detect_mode not in DEVICE_DETECT_MODES:
            msg = f"Invalid argument {device_detect_mode=}. (expected one of {DEVICE_DETECT_MODES})"
            raise ValueError(msg)

        super().__init__()
        self.__device_detect_mode: DeviceDetectMode = device_detect_mode
        if device_detect_mode == "proxy":
            self.register_buffer("__proxy", torch.empty((0,)), persistent=False)

    @property
    def device_detect_mode(self) -> DeviceDetectMode:
        return self.__device_detect_mode

    def get_device(self) -> Optional[torch.device]:
        """Returns the Module device according to device_detect_mode property."""
        if self.__device_detect_mode == "proxy":
            return self._buffers["__proxy"].device  # type: ignore
        elif self.__device_detect_mode == "first_param":
            try:
                device0 = next(self._get_devices_iterator(params=True, buffers=True))
                return device0
            except StopIteration:
                return None
        else:
            return None

    def get_devices(
        self,
        *,
        params: bool = True,
        buffers: bool = True,
        recurse: bool = True,
        output_type: Callable[[Iterator[torch.device]], T] = list,
    ) -> T:
        return output_type(
            self._get_devices_iterator(
                params=params,
                buffers=buffers,
                recurse=recurse,
            )
        )

    def _get_devices_iterator(
        self,
        *,
        params: bool = True,
        buffers: bool = True,
        recurse: bool = True,
    ) -> Iterator[torch.device]:
        """Returns an iterator over all unique devices in module."""
        its: List[Iterator[Union[Tensor, Parameter]]] = []
        if params:
            its.append(self.parameters(recurse=recurse))
        if buffers:
            its.append(self.buffers(recurse=recurse))

        devices = {}
        for it in its:
            for param_or_buffer in it:
                device = param_or_buffer.device
                if device not in devices:
                    yield device
                devices[param_or_buffer.device] = None


class ConfigModule(Generic[T_MutableMappingStr], nn.Module):
    _CONFIG_TYPES: ClassVar[Tuple[type, ...]] = (int, str, bool, float, NoneType)
    _CONFIG_EXCLUDE: ClassVar[Tuple[str, ...]] = ("^_.*",) + tuple(
        f".*{k}$" for k in nn.Module().__dict__.keys()
    )

    def __init__(
        self,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        config: Optional[T_MutableMappingStr] = None,
    ) -> None:
        if config is None:
            config = {}  # type: ignore

        attrs = {
            "config": config,
            "strict_load": strict_load,
            "config_to_extra_repr": config_to_extra_repr,
        }
        for name, value in attrs.items():
            object.__setattr__(self, f"_{ConfigModule.__name__}__{name}", value)

        super().__init__()
        self.__config: T_MutableMappingStr
        self.__strict_load: bool
        self.__config_to_extra_repr: bool

    @property
    def config(self) -> T_MutableMappingStr:
        return self.__config

    def __setattr__(self, name: str, value: Any) -> None:
        self.__update_config(name, value)
        return super().__setattr__(name, value)

    def __delattr__(self, name) -> None:
        self.__config.pop(name, None)
        return super().__delattr__(name)

    def extra_repr(self) -> str:
        if not self.__config_to_extra_repr:
            return super().extra_repr()
        else:
            return dump_dict(self.config)

    def add_module(self, name: str, module: Union[nn.Module, None]) -> None:
        self.__update_config(name, module)
        return super().add_module(name, module)

    def get_extra_state(self) -> Any:
        state = {"config": self.__config}
        return state

    def set_extra_state(self, state: Any) -> None:
        if not self.__strict_load:
            return None

        in_config = state["config"]
        if self.config == in_config:
            return None

        if is_dict_str(in_config) and is_dict_str(self.config):
            MISSING = "<missing>"
            union = set(in_config.keys()).union(self.config.keys())
            msgs = []
            for key in union:
                v1 = in_config.get(key, MISSING)
                v2 = in_config.get(key, MISSING)
                if v1 != v2:
                    msgs.append(f"{v1} != {v2}")
            msg = (
                "Invalid loaded config with current one. Invalid keys are:\n"
                + "\n\t".join(msgs)
            )
        else:
            msg = f"Invalid loaded config {in_config} with current one {self.config}."

        raise ValueError(msg)

    def __update_config(self, name: str, value: Any) -> None:
        subconfig = self.__class__._detect_subconfig(name, value)
        self.__config.update(subconfig)

    @classmethod
    def _detect_subconfig(cls, name: str, value: Any) -> Dict[str, Any]:
        prefix = f"{name}."
        if cls._is_config_name_value(name, value):
            subconfig = {name: value}
            prefix = ""
        elif isinstance(value, ConfigModule):
            subconfig = value.config
        elif hasattr(value, "_hparams") and is_mapping_str(value._hparams):
            subconfig = dict(value._hparams.items())  # type: ignore
        elif isinstance(value, torch.nn.Module):
            subconfig = cls._detect_torch_module_subconfig(value)
        elif hasattr(value, "__dict__"):
            subconfig = value.__dict__
        else:
            subconfig = {}

        subconfig = {f"{prefix}{k}": v for k, v in subconfig.items()}
        subconfig = {
            k: v for k, v in subconfig.items() if cls._is_config_name_value(k, v)
        }
        return subconfig

    @classmethod
    def _detect_torch_module_subconfig(cls, value: torch.nn.Module) -> Dict[str, Any]:
        subconfig = {
            k: v
            for k, v in value.__dict__.items()
            if k != "_modules" and match_patterns(k, exclude=cls._CONFIG_EXCLUDE)
        }
        subconfig.update(
            {
                kv: vv
                for k, v in value.__dict__["_modules"].items()
                for kv, vv in cls._detect_subconfig(k, v).items()
            }
        )
        return subconfig

    @classmethod
    def _is_config_name_value(cls, name: str, value: Any) -> bool:
        if not match_patterns(name, exclude=cls._CONFIG_EXCLUDE):
            return False
        else:
            return cls._is_config_value(value)

    @classmethod
    def _is_config_value(cls, value) -> bool:
        if isinstance(value, cls._CONFIG_TYPES):
            return True
        elif isinstance(value, (list, tuple, set, frozenset)):
            return all(cls._is_config_value(vi) for vi in value)
        elif isinstance(value, dict):
            return all(
                cls._is_config_value(k) and cls._is_config_value(v)
                for k, v in value.items()
            )
        else:
            return False


class TypedModule(Generic[InType, OutType], nn.Module):
    """Typed version of torch.nn.Module. Can specify an input and output type."""

    def __call__(self, *args: InType, **kwargs: InType) -> OutType:
        return super().__call__(*args, **kwargs)

    @overload
    def chain(
        self,
        *others: TypedModuleLike[Any, OutType],
    ) -> "ESequential[InType, OutType]":
        ...

    @overload
    def chain(self, *others: nn.Module) -> "ESequential[InType, Any]":
        ...

    def chain(self, *others):
        return ESequential(self, *others)

    def __or__(
        self,
        other: TypedModuleLike[Any, OutType],
    ) -> "ESequential[InType, OutType]":
        return self.chain(other)

    def __ror__(
        self,
        other: TypedModuleLike[InType, Any],
    ) -> "ESequential[InType, OutType]":
        return ESequential(other, self)


class TypedSequential(
    Generic[InType, OutType],
    TypedModule[InType, OutType],
    nn.Sequential,
):
    def __init__(
        self,
        *args,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
    ) -> None:
        TypedModule.__init__(self)
        nn.Sequential.__init__(self, *args)

        self.__unpack_tuple = unpack_tuple
        self.__unpack_dict = unpack_dict

    @property
    def unpack_tuple(self) -> bool:
        return self.__unpack_tuple

    @property
    def unpack_dict(self) -> bool:
        return self.__unpack_dict

    def __call__(self, x: InType) -> OutType:  # type: ignore
        return nn.Sequential.__call__(self, x)

    def forward(self, x: InType) -> OutType:  # type: ignore
        for module in self:
            if self.__unpack_tuple and isinstance(x, tuple):
                x = module(*x)
            elif self.__unpack_dict and is_dict_str(x):
                x = module(**x)
            else:
                x = module(x)
        return x  # type: ignore

    def tolist(self) -> List[nn.Module]:
        return list(self._modules.values())

    def todict(self) -> Dict[str, nn.Module]:
        return copy.copy(self._modules)


class EModule(
    Generic[InType, OutType],
    ConfigModule,
    TypedModule[InType, OutType],
    ProxyDeviceModule,
):
    """Enriched torch.nn.Module with proxy device, forward typing and automatic configuration detection from attributes.

    The default behaviour is the same than PyTorch Module class.
    """

    def __init__(
        self,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        """
        Args:
            strict_load: If True, Module config will be compared during load_state_dict(...) method call and raises a ValueError. defaults to False.
            config_to_extra_repr: If True, add config to extra repr. defaults to False.
            device_detect_mode: Enable automatic detection of the module device. defaults to "first_param".
        """
        # ConfigModule must be first
        ConfigModule.__init__(
            self,
            strict_load=strict_load,
            config_to_extra_repr=config_to_extra_repr,
        )
        TypedModule.__init__(self)
        ProxyDeviceModule.__init__(
            self,
            device_detect_mode=device_detect_mode,
        )

    def count_parameters(
        self,
        *,
        recurse: bool = True,
        only_trainable: bool = False,
        buffers: bool = False,
    ) -> int:
        """Returns the number of parameters in this module."""
        return count_parameters(
            self,
            recurse=recurse,
            only_trainable=only_trainable,
            buffers=buffers,
        )

    def checksum(
        self,
        *,
        only_trainable: bool = False,
        with_names: bool = False,
        buffers: bool = False,
        training: bool = False,
    ) -> int:
        return checksum_module(
            self,
            only_trainable=only_trainable,
            with_names=with_names,
            buffers=buffers,
            training=training,
        )


class ESequential(
    Generic[InType, OutType],
    EModule[InType, OutType],
    TypedSequential[InType, OutType],
):
    """Enriched torch.nn.Sequential with proxy device, forward typing and automatic configuration detection from attributes.

    Designed to work with `torchoutil.nn.EModule` instances.
    The default behaviour is the same than PyTorch Sequential class.
    """

    @overload
    def __init__(
        self,
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
        unpack_dict: bool = False,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, Any],
        arg4: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, Any],
        arg4: TypedModuleLike[Any, Any],
        arg5: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TypedModuleLike[InType, Any],
        arg1: TypedModuleLike[Any, Any],
        arg2: TypedModuleLike[Any, Any],
        arg3: TypedModuleLike[Any, Any],
        arg4: TypedModuleLike[Any, Any],
        arg5: TypedModuleLike[Any, Any],
        arg6: TypedModuleLike[Any, OutType],
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, TypedModuleLike[InType, OutType]]",
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, nn.Module]",
        /,
        *,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *args: nn.Module,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    def __init__(
        self,
        *args,
        unpack_tuple: bool = False,
        unpack_dict: bool = False,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        """
        Args:
            unpack_tuple: If True, the outputs of a module that returns a tuple at position i will be unpacked for positional arguments for the next module at position i+1. defaults to False.
            unpack_tuple: If True, the outputs of a module that returns a dict at position i will be unpacked for keywords arguments for the next module at position i+1. defaults to False.
            strict_load: If True, Module config will be compared during load_state_dict(...) method call and raises a ValueError. defaults to False.
            config_to_extra_repr: If True, add config to extra repr. defaults to False.
            device_detect_mode: Enable automatic detection of the module device. defaults to "first_param".
        """
        EModule.__init__(
            self,
            strict_load=strict_load,
            config_to_extra_repr=config_to_extra_repr,
            device_detect_mode=device_detect_mode,
        )
        TypedSequential.__init__(
            self,
            *args,
            unpack_tuple=unpack_tuple,
            unpack_dict=unpack_dict,
        )


class EModuleList(
    Generic[InType, OutType3],
    EModule[InType, List[OutType3]],
    nn.ModuleList,
):
    """Enriched torch.nn.ModuleList with proxy device, forward typing and automatic configuration detection from attributes.

    Designed to work with `torchoutil.nn.EModule` instances.
    The default behaviour is the same than PyTorch ModuleList class, except for the forward call which returns a list containing the output of each module called separately.
    """

    @overload
    def __init__(
        self,
        modules: Optional[Iterable[TypedModuleLike[InType, OutType3]]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        modules: Optional[Iterable[nn.Module]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    def __init__(
        self,
        modules=None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        EModule.__init__(
            self,
            strict_load=strict_load,
            config_to_extra_repr=config_to_extra_repr,
            device_detect_mode=device_detect_mode,
        )
        nn.ModuleList.__init__(self, modules)

    def forward(self, *args: InType, **kwargs: InType) -> List[OutType3]:
        return [module(*args, **kwargs) for module in self]


class EModuleDict(
    Generic[InType, OutType3],
    EModule[InType, Dict[str, OutType3]],
    nn.ModuleDict,
):
    """Enriched torch.nn.ModuleDict with proxy device, forward typing and automatic configuration detection from attributes.

    Designed to work with `torchoutil.nn.EModule` instances.
    The default behaviour is the same than PyTorch ModuleDict class, except for the forward call which returns a dict containing the output of each module called separately.
    """

    @overload
    def __init__(
        self,
        modules: Optional[Mapping[str, TypedModuleLike[InType, OutType3]]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        modules: Optional[Mapping[str, nn.Module]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        ...

    def __init__(
        self,
        modules=None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        EModule.__init__(
            self,
            strict_load=strict_load,
            config_to_extra_repr=config_to_extra_repr,
            device_detect_mode=device_detect_mode,
        )
        nn.ModuleDict.__init__(self, modules)

    def forward(self, *args: InType, **kwargs: InType) -> Dict[str, OutType3]:
        return {name: module(*args, **kwargs) for name, module in self.items()}


class EModulePartial(
    Generic[InType, OutType],
    EModule[InType, OutType],
):
    def __init__(
        self,
        fn: Callable[Concatenate[InType, P], OutType],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def forward(self, x: InType) -> OutType:  # type: ignore
        return self.fn(x, *self.args, **self.kwargs)

    def extra_repr(self) -> str:
        return f"{self.fn.__name__}, {ConfigModule.extra_repr(self)}"


def __test_typing_1() -> None:
    import torch
    from torch import Tensor

    class LayerA(TypedModule[Tensor, Tensor]):
        def forward(self, x: Tensor) -> Tensor:
            return x * x

    class LayerB(TypedModule[Tensor, int]):
        def forward(self, x: Tensor) -> int:
            return int(x.sum().item())

    class LayerC(TypedModule[int, Tensor]):
        def forward(self, x: int) -> Tensor:
            return torch.as_tensor(x)

    x = torch.rand(10)
    xa = LayerA()(x)
    xb = LayerB()(x)

    seq = ESequential(LayerA(), LayerA(), LayerB())
    xab = seq(x)

    seq = LayerA() | LayerA() | LayerB()
    xab = seq(x)

    seq = LayerC().chain(LayerA())
    xc = seq(2)

    assert isinstance(xa, Tensor)
    assert isinstance(xb, int)
    assert isinstance(xab, int)
    assert isinstance(xc, Tensor)

    class LayerD(nn.Module):
        def forward(self, x: Tensor) -> int:
            return int(x.item())

    class LayerE(nn.Module):
        def forward(self, x: bool) -> str:
            return str(x)

    seq = ESequential(LayerD(), LayerE())
    y = seq(torch.rand())

    assert isinstance(y, str)

    class LayerF(TypedModule[bool, str]):
        def forward(self, x):
            return str(x)

    seq = ESequential(LayerF())
    y = seq(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import sys
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from typing_extensions import override

from torchoutil import Tensor
from torchoutil.pyoutil import is_iterable_str
from torchoutil.types.tensor_subclasses import Tensor1D
from torchoutil.utils.data import DatasetSlicer
from torchoutil.utils.pack.common import (
    ATTRS_FNAME,
    ContentMode,
    ItemType,
    PackedDatasetAttributes,
    _dict_to_tuple,
)
from torchoutil.utils.saving.load_fn import LOAD_FNS, LoadFn, LoadFnLike

T = TypeVar("T")
U = TypeVar("U")

pylog = logging.getLogger(__name__)


class PackedDataset(Generic[T, U], DatasetSlicer[U]):
    def __init__(
        self,
        root: Union[str, Path],
        transform: Optional[Callable[[T], U]] = None,
        *,
        load_fn: LoadFnLike[Union[T, List[T]]] = "pickle",
    ) -> None:
        """

        Args:
            root: Root directory containing the info.json and the data files.
            transform: Optional transform to apply to each item. defaults to None.
            load_fn: Load function to load an item or batch. defaults to torch.load.
        """
        root = Path(root)
        if isinstance(load_fn, str):
            load_fn = LOAD_FNS[load_fn]

        super().__init__()
        self._root = root
        self._transform = transform
        self._load_fn = load_fn

        self._attrs: PackedDatasetAttributes = {}  # type: ignore
        self._fpaths: List[Path] = []
        self._column_to_fname: Dict[str, str] = {}
        self._reload_data(root, load_fn)

    @property
    def attrs(self) -> Optional[PackedDatasetAttributes]:
        return self._attrs  # type: ignore

    @property
    def content_mode(self) -> Optional[ContentMode]:
        return self._attrs.get("content_mode")

    @property
    def item_type(self) -> ItemType:
        return self._attrs.get("item_type", "raw")

    @property
    def batch_size(self) -> int:
        return self._attrs["batch_size"]

    @override
    def __len__(self) -> int:
        return self._attrs["length"]

    def to_dict(self) -> Dict[str, Sequence]:
        if self.item_type != "dict":
            msg = f"Cannot convert non-dict dataset to dict. (found {self.item_type=})"
            raise ValueError(msg)

        if self.content_mode == "column":
            return self._load_item_from_columns(slice(None), None)
        else:
            return self[:]  # type: ignore

    @override
    def get_item(self, idx: int) -> Union[T, U]:
        if self.content_mode == "column":
            item = self._load_item_from_columns(idx)
            if self._transform is not None:
                item = self._transform(item)  # type: ignore
            return item
        elif self.content_mode == "item":
            batch_size = 1
        elif self.content_mode == "batch":
            batch_size = self.batch_size
        else:
            msg = f"Invalid PickleDataset state. (cannot load item with {self.content_mode=})"
            raise RuntimeError(msg)

        target_idx = idx // batch_size
        fpath = self._fpaths[target_idx]
        item_or_batch = self._load_fn(fpath)

        if self.content_mode == "item":
            item: T = item_or_batch  # type: ignore
        else:
            batch: List[T] = item_or_batch  # type: ignore
            local_idx = idx % batch_size
            item = batch[local_idx]

        if self._transform is not None:
            item = self._transform(item)  # type: ignore
        return item  # type: ignore

    @override
    def get_items_indices(
        self,
        indices: Union[Iterable[int], Tensor1D],
        *args,
    ) -> List[U]:
        if self.content_mode == "column":
            return self._load_item_from_columns(indices)
        else:
            return super().get_items_indices(indices, *args)

    @override
    def get_items_mask(
        self,
        mask: Union[Iterable[bool], Tensor],
        *args,
    ) -> List[U]:
        if self.content_mode == "column":
            return self._load_item_from_columns(mask)  # type: ignore
        else:
            return super().get_items_mask(mask, *args)

    @override
    def get_items_slice(
        self,
        slice_: slice,
        *args,
    ) -> List[U]:
        if self.content_mode == "column":
            return self._load_item_from_columns(slice_)
        else:
            return super().get_items_slice(slice_, *args)

    def _load_item_from_columns(
        self,
        idx: Union[int, Iterable[int], Iterable[bool], Tensor1D, slice],
        column: Union[str, Iterable[str], None] = None,
    ) -> Any:
        if isinstance(column, str):
            columns = [column]
        elif is_iterable_str(column):
            columns = column
        elif column is None:
            columns = self._column_to_fname.keys()
        else:
            raise TypeError(f"Invalid argument type {type(column)=}.")

        fnames = dict.fromkeys(self._column_to_fname[column_i] for column_i in columns)
        fpaths = [
            fpath for fname in fnames for fpath in self._fpaths if fpath.name == fname
        ]
        loaded = {fpath.name: self._load_fn(fpath) for fpath in fpaths}
        fname_to_column = {
            fname: column for column, fname in self._column_to_fname.items()
        }

        item = {fname_to_column[fname]: v[idx] for fname, v in loaded.items()}  # type: ignore
        if self.item_type == "tuple":
            item = _dict_to_tuple(item)  # type: ignore
        return item

    def _reload_data(
        self,
        root: Path,
        load_fn: LoadFn[Union[T, List[T]]],
    ) -> None:
        attrs_fpath = root.joinpath(ATTRS_FNAME)

        if not attrs_fpath.is_file():
            raise FileNotFoundError(f"Cannot find attribute file '{str(attrs_fpath)}'.")

        with open(attrs_fpath, "r") as file:
            attrs = json.load(file)

        # Disable check for python <= 3.8 because __required_keys__ does not exists in this version
        if sys.version_info.major < 3 or (
            sys.version_info.major == 3 and sys.version_info.minor <= 8
        ):
            missing = []
        else:
            missing = list(
                set(PackedDatasetAttributes.__required_keys__).difference(attrs)
            )

        if len(missing) > 0:
            msg = f"Missing {len(missing)} keys in attribute file. (with {missing=} in {attrs_fpath=})"
            raise RuntimeError(msg)

        content_dname = attrs["content_dname"]
        content_dpath = root.joinpath(content_dname)

        fnames = attrs["files"]
        fpaths = [content_dpath.joinpath(fname) for fname in fnames]
        column_to_fname = attrs.get("column_to_fname", {})

        self._root = root
        self._load_fn = load_fn
        self._attrs = attrs
        self._fpaths = fpaths
        self._column_to_fname = column_to_fname

    @classmethod
    def is_packed_root(cls, root: Union[str, Path]) -> bool:
        try:
            PackedDataset(root)
            return True
        except (FileNotFoundError, RuntimeError):
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

import pyoutil as po
from torchoutil.core.packaging import _NUMPY_AVAILABLE
from torchoutil.extras.numpy import np
from torchoutil.nn.functional.transform import (
    flatten,
    repeat_interleave_nd,
    resample_nearest_rates,
)


class TestRepeat(TestCase):
    def test_example_1(self) -> None:
        x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        result = repeat_interleave_nd(x, repeats=2, dim=0)
        expected = torch.as_tensor(
            [[0, 1, 2, 3], [0, 1, 2, 3], [4, 5, 6, 7], [4, 5, 6, 7]]
        )
        assert torch.equal(result, expected)


class TestResampleNearest(TestCase):
    def test_example_1(self) -> None:
        x = torch.arange(10, 20)
        result = resample_nearest_rates(x, 0.5)
        expected = torch.as_tensor([10, 12, 14, 16, 18])
        assert torch.equal(result, expected)

    def test_example_2(self) -> None:
        x = torch.arange(10, 20)
        result = resample_nearest_rates(x, 2)
        expected = torch.as_tensor(
            [
                10,
                10,
                11,
                11,
                12,
                12,
                13,
                13,
                14,
                14,
                15,
                15,
                16,
                16,
                17,
                17,
                18,
                18,
                19,
                19,
            ]
        )
        assert torch.equal(result, expected)

    def test_example_3(self) -> None:
        x = torch.stack([torch.arange(10, 20), torch.arange(20, 30)])
        result = resample_nearest_rates(x, 0.5)
        expected = torch.as_tensor([[10, 12, 14, 16, 18], [20, 22, 24, 26, 28]])
        assert torch.equal(result, expected)


class TestFlatten(TestCase):
    def test_example_1(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        x = [
            [
                [[np.float32(64), 3.0, 0, 1], ["a", None, "", 2], torch.zeros(4)],
                torch.ones(3, 4),
            ],
        ]
        expected = [
            [np.float32(64), 3.0, 0, 1, "a", None, "", 2]
            + list(torch.zeros(4))
            + list(torch.ones(12))
        ]

        assert len(x) == len(expected)
        for xi, expected_i in zip(x, expected):
            result_i = flatten(xi)
            assert result_i == expected_i

    def test_example_2_between_dims(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        shape = (10, 3, 4, 5)

        x = [
            (np.zeros(shape), 0, len(shape)),
            (np.zeros(shape), 1, 2),
            (np.zeros(shape), 0, 1),
            (np.float64(1.0), 0, None),
        ]
        expected = [
            np.zeros(po.prod(shape)),
            np.zeros((10, 12, 5)),
            np.zeros((30, 4, 5)),
            np.ones(1, dtype=np.float64),
        ]

        assert len(x) == len(expected)
        for (xi, start, end), expected_i in zip(x, expected):
            result_i = flatten(xi, start, end)
            assert np.equal(result_i, expected_i).all()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.modules import (
    Abs,
    Angle,
    AsTensor,
    CropDim,
    CropDims,
    Identity,
    LogSoftmaxMultidim,
    Mean,
    PadAndStackRec,
    PadDim,
    PadDims,
    Permute,
    RepeatInterleaveNd,
    ResampleNearestRates,
    SoftmaxMultidim,
    ToList,
    Transpose,
    Unsqueeze,
)
from torchoutil.nn.modules.mixins import ESequential


class TestSequential(TestCase):
    def test_example_1(self) -> None:
        target_length = 10
        transform = ESequential(
            ResampleNearestRates(0.5),
            PadDim(target_length),
            ResampleNearestRates(2.5),
            CropDim(target_length),
        )

        x = torch.rand(10, 20, target_length)
        result = transform(x)

        assert x.shape == result.shape

    def test_example_2(self) -> None:
        transform = ESequential(
            ToList(),
            AsTensor(),
            Abs(),
            Angle(),
            PadAndStackRec(0.0),
            PadDims([10]),
            CropDims([10]),
            Mean(dim=1),
            Unsqueeze(dim=1),
            Permute(1, 0),
            RepeatInterleaveNd(10, 0),
            Transpose(0, 1),
            LogSoftmaxMultidim(dims=(0, 1)),
            SoftmaxMultidim(dims=(1,)),
        )

        x = torch.rand(16, 10)
        result = transform(x)

        assert x.shape == result.shape

    def test_example_3(self) -> None:
        transform = ESequential(
            ToList(),
            Identity(),
            AsTensor(),
        )

        x = torch.rand(16, 10)
        result = transform(x)

        assert x.shape == result.shape


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Callable, Dict, Literal, TypeVar, Union

import torch

from torchoutil.core.packaging import (
    _NUMPY_AVAILABLE,
    _SAFETENSORS_AVAILABLE,
    _YAML_AVAILABLE,
)
from torchoutil.extras.numpy import np
from torchoutil.utils.saving.csv import to_csv
from torchoutil.utils.saving.json import to_json
from torchoutil.utils.saving.pickle import to_pickle

T = TypeVar("T", covariant=True)

SaveFn = Callable[[T, Path], None]
SafeFnName = Literal["csv", "json", "numpy", "pickle", "safetensors", "torch", "yaml"]
SaveFnLike = Union[SaveFn[T], SafeFnName]


def numpy_save_fn(obj: Any, fpath: Path) -> None:
    np.save(fpath, obj)


SAVE_FNS: Dict[str, SaveFn[Any]] = {
    "csv": to_csv,  # type: ignore
    "json": to_json,
    "pickle": to_pickle,
    "torch": torch.save,
}
SAVE_EXTENSIONS = {
    "csv": "csv",
    "json": "json",
    "pickle": "pickle",
    "torch": "pt",
}

if _NUMPY_AVAILABLE:
    SAVE_FNS["numpy"] = numpy_save_fn
    SAVE_EXTENSIONS["numpy"] = "npy"

if _SAFETENSORS_AVAILABLE:
    from torchoutil.utils.saving.safetensors import to_safetensors

    SAVE_FNS["safetensors"] = to_safetensors
    SAVE_EXTENSIONS["safetensors"] = "safetensors"

if _YAML_AVAILABLE:
    from torchoutil.utils.saving.yaml import to_yaml

    SAVE_FNS["yaml"] = to_yaml  # type: ignore
    SAVE_EXTENSIONS["yaml"] = "yaml"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Dict, Optional, Union

from safetensors.torch import load_file, save_file
from torch import Tensor


def load_safetensors(
    fpath: Union[str, Path],
    *,
    device: str = "cpu",
) -> Dict[str, Tensor]:
    return load_file(fpath, device)


def to_safetensors(
    tensors: Dict[str, Tensor],
    fpath: Union[str, Path],
    metadata: Optional[Dict[str, str]] = None,
) -> None:
    return save_file(tensors, fpath, metadata)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.modules.mixins import ESequential
from torchoutil.nn.modules.multiclass import (
    IndexToName,
    IndexToOnehot,
    NameToIndex,
    NameToOnehot,
    OnehotToIndex,
    OnehotToName,
    ProbsToOnehot,
)


class TestMulticlass(TestCase):
    def test_example_1(self) -> None:
        for _ in range(10):
            num_steps = int(torch.randint(1, 10, ()).item())
            num_samples = int(torch.randint(0, 20, ()).item())
            num_classes = int(torch.randint(1, 20, ()).item())
            idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))

            probs = torch.rand(num_steps, num_samples, num_classes)
            onehot = ProbsToOnehot()(probs)

            # dummy pipeline to convert labels multiple times
            pipeline = ESequential(
                OnehotToName(idx_to_name),
                NameToIndex(idx_to_name),
                IndexToOnehot(len(idx_to_name)),
                OnehotToIndex(),
                IndexToName(idx_to_name),
                NameToOnehot(idx_to_name),
            )
            result = pipeline(onehot)

            assert torch.equal(onehot, result), f"{onehot=}; {result=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import unittest
from typing import Sized
from unittest import TestCase

import torch

from torchoutil.core.packaging import _NUMPY_AVAILABLE
from torchoutil.extras.numpy import np
from torchoutil.nn import functional as F
from torchoutil.pyoutil.typing import is_builtin_number, is_builtin_scalar
from torchoutil.types import (
    is_number_like,
    is_numpy_number_like,
    is_scalar_like,
    is_tensor0d,
)


class TestIsNumber(TestCase):
    def test_example_1(self) -> None:
        tests = [
            (1, True),
            (1.0, True),
            (False, True),
            (1j + 2, True),
            (torch.rand(()), True),
            (torch.rand((0,)), False),
            (torch.rand(10), False),
            ([], False),
            ([[]], False),
            ((), False),
            ("test", False),
            (None, False),
            ("", False),
            (b"abc", False),
            (bytearray(), False),
            ([1, [2]], False),
            ({}, False),
            ({"a": [1, 2], "b": [3, 4]}, False),
            (set(), False),
            ((1, 2, 3), False),
            ([1.0], False),
            ([object(), [], "abc"], False),
            (np.array(["a", "b"])[0], False),
        ]

        if _NUMPY_AVAILABLE:
            tests += [
                (np.float64(random.random()), True),
                (np.int64(random.randint(0, 100)), True),
                (np.random.random(()), True),
                (np.random.random((0,)), False),
                (np.random.random((10,)), False),
            ]

        for x, expected in tests:
            x_is_number = is_number_like(x)
            msg = f"{x=} ({is_builtin_number(x)}, {is_tensor0d(x)}, {is_numpy_number_like(x)})"
            assert x_is_number == expected, msg

            x_is_scalar = is_scalar_like(x)
            assert isinstance(x, Sized) or x_is_scalar

            # Impl: number => scalar
            assert not x_is_number or x_is_scalar

            if _NUMPY_AVAILABLE:
                np_x_is_scalar = np.isscalar(x)
                # Impl: np_scalar => scalar
                assert not np_x_is_scalar or x_is_scalar, f"{x=}"

            try:
                # Impl: scalar => (ndim == 0)
                ndim = F.ndim(x)  # type: ignore
                assert not x_is_scalar or (ndim == 0), f"{type(x)=} ; {x=}"

                # Impl: scalar => (len(shape) == 0)
                shape = F.shape(x)  # type: ignore
                assert not x_is_scalar or (len(shape) == 0), f"{type(x)=} ; {x=}"

                assert len(shape) == ndim, f"{type(x)=} ; {x=}"

                if _NUMPY_AVAILABLE:
                    np_x = np.array(x)
                    assert shape == np_x.shape, f"{x=}"

                # Impl: scalar => (nelements == 1)
                nelements = F.nelement(x)  # type: ignore
                assert not x_is_scalar or (nelements == 1), f"{type(x)=} ; {x=}"

                xitem = F.to_item(x)  # type: ignore
                assert is_builtin_scalar(xitem), f"{x=}"

            except (ValueError, RuntimeError, TypeError) as err:
                assert not x_is_scalar, f"{type(x)=} ; {x=} ; {err=}"

    def test_is_builtin_strict(self) -> None:
        class subint(int):
            ...

        i1 = 100
        i2 = subint(100)

        assert is_builtin_scalar(i1, strict=False)
        assert is_builtin_scalar(i1, strict=True)
        assert is_builtin_scalar(i2, strict=False)
        assert not is_builtin_scalar(i2, strict=True)

        if not _NUMPY_AVAILABLE:
            return None

        s1 = "abc"
        arr = np.array([s1, s1])
        s2 = arr[0]

        assert is_builtin_scalar(s1, strict=False)
        assert is_builtin_scalar(s1, strict=True)
        assert is_builtin_scalar(s2, strict=False)
        assert not is_builtin_scalar(s2, strict=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple, Union

import torch
from torch import Tensor

from torchoutil.nn.functional.pad import pad_dim
from torchoutil.types import BoolTensor, LongTensor


def extract_segments(x: Tensor) -> LongTensor:
    """
    Example 1
    ----------
    >>> x = torch.as_tensor([0, 1, 1, 0, 0, 1, 1, 1, 1, 0]).bool()
    >>> starts, ends = extract_segments(x)
    >>> starts
    ... tensor([1, 5])
    >>> ends
    ... tensor([3, 9])

    Example 2
    ----------
    >>> x = torch.as_tensor([[1, 1, 1, 0], [1, 0, 0, 1]]).bool()
    >>> indices, starts, ends = extract_segments(x)
    >>> indices
    ... tensor([0, 1, 1])
    >>> starts
    ... tensor([0, 0, 3])
    >>> ends
    ... tensor([3, 1, 4])

    Args:
        x: (..., N) bool tensor containing D dims

    Returns:
        segments: (D+1, M) tensor, where M is the total number of segments
            When D > 1, segments also contains indices of the source column for each start and end value. See Example 2 for detail.
    """
    if not isinstance(x, BoolTensor):
        raise ValueError(f"Invalid argument {x=}. (expected BoolTensor)")

    x = x.int()
    x = pad_dim(x, x.shape[-1] + 2, align="center", pad_value=0, dim=-1)
    diff = x[..., 1:] - x[..., :-1]

    starts = torch.stack(torch.where(diff > 0))
    ends = torch.stack(torch.where(diff < 0))

    result = torch.cat([starts, ends[-1:]], dim=0)
    return result  # type: ignore


def segments_to_list(
    segments: Tensor,
    maxsize: Union[int, Tuple[int, ...], None],
) -> list:
    if segments.shape[0] == 2:
        starts, ends = segments.tolist()
        return list(zip(starts, ends))

    elif segments.shape[0] in (0, 1):
        msg = f"Invalid argument shape {segments.shape=}. (expected first dim >= 2)"
        raise ValueError(msg)

    if maxsize is None:
        num_elems = segments[0].max().item() + 1
        next_maxsize = None
    elif isinstance(maxsize, tuple):
        num_elems = maxsize[0]
        next_maxsize = maxsize[1:]
    else:
        num_elems = maxsize
        next_maxsize = None

    indices = torch.arange(num_elems)
    result = [
        segments_to_list(segments[1:, ..., segments[0] == idx], maxsize=next_maxsize)
        for idx in indices
    ]
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import TypeVar, Union

import torch

from torchoutil.extras.numpy.definitions import (  # noqa: F401
    NumpyNumberLike,
    NumpyScalarLike,
    np,
    numpy,
)
from torchoutil.pyoutil.typing.classes import BuiltinNumber, BuiltinScalar

from .tensor_subclasses import (  # noqa: F401
    BoolTensor,
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    ByteTensor,
    ByteTensor0D,
    ByteTensor1D,
    ByteTensor2D,
    ByteTensor3D,
    CDoubleTensor,
    CDoubleTensor0D,
    CDoubleTensor1D,
    CDoubleTensor2D,
    CDoubleTensor3D,
    CFloatTensor,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    CHalfTensor,
    CHalfTensor0D,
    CHalfTensor1D,
    CHalfTensor2D,
    CHalfTensor3D,
    CharTensor,
    CharTensor0D,
    CharTensor1D,
    CharTensor2D,
    CharTensor3D,
    ComplexFloatingTensor,
    ComplexFloatingTensor0D,
    ComplexFloatingTensor1D,
    ComplexFloatingTensor2D,
    ComplexFloatingTensor3D,
    DoubleTensor,
    DoubleTensor0D,
    DoubleTensor1D,
    DoubleTensor2D,
    DoubleTensor3D,
    FloatingTensor,
    FloatingTensor0D,
    FloatingTensor1D,
    FloatingTensor2D,
    FloatingTensor3D,
    FloatTensor,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    HalfTensor,
    HalfTensor0D,
    HalfTensor1D,
    HalfTensor2D,
    HalfTensor3D,
    IntTensor,
    IntTensor0D,
    IntTensor1D,
    IntTensor2D,
    IntTensor3D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    ShortTensor,
    ShortTensor0D,
    ShortTensor1D,
    ShortTensor2D,
    ShortTensor3D,
    SignedIntegerTensor,
    SignedIntegerTensor0D,
    SignedIntegerTensor1D,
    SignedIntegerTensor2D,
    SignedIntegerTensor3D,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
    UnsignedIntegerTensor,
    UnsignedIntegerTensor0D,
    UnsignedIntegerTensor1D,
    UnsignedIntegerTensor2D,
    UnsignedIntegerTensor3D,
)

r"""/!\ The following type hints are meant for type annotation only, not for runtime checks.
"""

NumberLike = Union[BuiltinNumber, NumpyNumberLike, Tensor0D]
ScalarLike = Union[BuiltinScalar, NumpyScalarLike, Tensor0D]
TensorLike = Union[torch.Tensor, np.ndarray]

T_TensorLike = TypeVar(
    "T_TensorLike",
    bound=Union[torch.Tensor, np.ndarray],
    covariant=True,
)

del np



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import Namespace
from dataclasses import asdict
from enum import Enum
from pathlib import Path
from re import Pattern
from typing import Any, Dict, Iterable, List, Literal, Mapping, TypeVar, Union, overload

import torch
from torch import Tensor

from torchoutil.core.packaging import (
    _NUMPY_AVAILABLE,
    _OMEGACONF_AVAILABLE,
    _PANDAS_AVAILABLE,
)
from torchoutil.pyoutil.typing import (
    BuiltinNumber,
    DataclassInstance,
    NamedTupleInstance,
    T_BuiltinScalar,
    is_builtin_scalar,
    is_dataclass_instance,
    is_namedtuple_instance,
)

if _NUMPY_AVAILABLE:
    import numpy as np

if _OMEGACONF_AVAILABLE:
    from omegaconf import DictConfig, ListConfig, OmegaConf  # type: ignore

if _PANDAS_AVAILABLE:
    from pandas import DataFrame


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

UnkMode = Literal["identity", "error"]
UNK_MODES = ("identity", "error")


@overload
def to_builtin(
    x: T_BuiltinScalar,
    *,
    unk_mode: UnkMode = "error",
) -> T_BuiltinScalar:
    ...


@overload
def to_builtin(
    x: Enum,
    *,
    unk_mode: UnkMode = "error",
) -> str:
    ...


@overload
def to_builtin(
    x: Path,
    *,
    unk_mode: UnkMode = "error",
) -> str:
    ...


@overload
def to_builtin(
    x: Pattern,
    *,
    unk_mode: UnkMode = "error",
) -> str:
    ...


@overload
def to_builtin(
    x: Namespace,
    *,
    unk_mode: UnkMode = "error",
) -> Dict[str, Any]:
    ...


@overload
def to_builtin(
    x: Tensor,
    *,
    unk_mode: UnkMode = "error",
) -> Union[List, BuiltinNumber]:
    ...


@overload
def to_builtin(
    x: Mapping[K, V],
    *,
    unk_mode: UnkMode = "error",
) -> Dict[K, V]:
    ...


@overload
def to_builtin(
    x: DataclassInstance,
    *,
    unk_mode: UnkMode = "error",
) -> Dict[str, Any]:
    ...


@overload
def to_builtin(
    x: NamedTupleInstance,
    *,
    unk_mode: UnkMode = "error",
) -> Dict[str, Any]:
    ...


@overload
def to_builtin(
    x: Any,
    *,
    unk_mode: UnkMode = "error",
) -> Any:
    ...


def to_builtin(
    x: Any,
    *,
    unk_mode: UnkMode = "error",
) -> Any:
    """Helper function to sanitize data before saving to YAML or CSV file.

    Args:
        x: Object to convert to built-in equivalent.
        unk_mode: When an object type is not recognized, unk_mode defines the behaviour.
            If unk_mode == "identity", the object is returned unchanged.
            If unk_mode == "error", a TypeError is raised.
    """
    # Terminal cases
    if is_builtin_scalar(x, strict=True):
        return x
    if isinstance(x, Enum):
        return x.name
    if isinstance(x, Path):
        return str(x)
    if isinstance(x, Pattern):
        return x.pattern
    if isinstance(x, Tensor):
        return x.tolist()
    if isinstance(x, torch.dtype):
        return str(x)
    if _NUMPY_AVAILABLE and isinstance(x, np.ndarray):  # type: ignore
        return x.tolist()
    if _NUMPY_AVAILABLE and isinstance(x, np.generic):  # type: ignore
        return x.item()
    if _NUMPY_AVAILABLE and isinstance(x, np.dtype):  # type: ignore
        return str(x)

    # Non-terminal cases (iterables and mappings)
    if _OMEGACONF_AVAILABLE and isinstance(x, (DictConfig, ListConfig)):  # type: ignore
        return to_builtin(OmegaConf.to_container(x, resolve=False, enum_to_str=True))  # type: ignore
    if _PANDAS_AVAILABLE and isinstance(x, DataFrame):  # type: ignore
        return to_builtin(x.to_dict("list"))
    if isinstance(x, Namespace):
        return to_builtin(x.__dict__)
    if is_dataclass_instance(x):
        return to_builtin(asdict(x))
    if is_namedtuple_instance(x):
        return to_builtin(x._asdict())
    if isinstance(x, Mapping):
        return {to_builtin(k): to_builtin(v) for k, v in x.items()}
    if isinstance(x, Iterable):
        return [to_builtin(xi) for xi in x]

    if unk_mode == "identity":
        return x
    elif unk_mode == "error":
        raise TypeError(f"Unsupported argument type {type(x)}.")
    else:
        raise ValueError(f"Invalid argument {unk_mode=}. (expected one of {UNK_MODES})")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Generic, List, Mapping, Optional, Union

import torch
from torch import Tensor, nn

from torchoutil.core.get import DeviceLike, DTypeLike
from torchoutil.nn.functional.multilabel import (
    T_Name,
    indices_to_multihot,
    indices_to_multinames,
    multihot_to_indices,
    multihot_to_multinames,
    multinames_to_indices,
    multinames_to_multihot,
    probs_to_indices,
    probs_to_multihot,
    probs_to_multinames,
)
from torchoutil.pyoutil.collections import dump_dict


class IndicesToMultihot(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.indices_to_multihot`.
    """

    def __init__(
        self,
        num_classes: int,
        *,
        padding_idx: Optional[int] = None,
        device: DeviceLike = None,
        dtype: DTypeLike = torch.bool,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.padding_idx = padding_idx
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        indices: Union[List[List[int]], List[Tensor]],
    ) -> Tensor:
        multihot = indices_to_multihot(
            indices,
            self.num_classes,
            padding_idx=self.padding_idx,
            device=self.device,
            dtype=self.dtype,
        )
        return multihot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                num_classes=self.num_classes,
                padding_idx=self.padding_idx,
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class IndicesToMultinames(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.indices_to_multinames`.
    """

    def __init__(
        self,
        idx_to_name: Mapping[int, T_Name],
        *,
        padding_idx: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name
        self.padding_idx = padding_idx

    def forward(
        self,
        indices: Union[List[List[int]], List[Tensor]],
    ) -> List[List[T_Name]]:
        names = indices_to_multinames(
            indices,
            self.idx_to_name,
            padding_idx=self.padding_idx,
        )
        return names

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                padding_idx=self.padding_idx,
            ),
            ignore_lst=(None,),
        )


class MultihotToIndices(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.multihot_to_indices`.
    """

    def __init__(self, *, padding_idx: Optional[int] = None) -> None:
        super().__init__()
        self.padding_idx = padding_idx

    def forward(
        self,
        multihot: Tensor,
    ) -> List[List[int]]:
        names = multihot_to_indices(multihot, padding_idx=self.padding_idx)
        return names

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                padding_idx=self.padding_idx,
            ),
            ignore_lst=(None,),
        )


class MultihotToMultinames(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.multihot_to_multinames`.
    """

    def __init__(
        self,
        idx_to_name: Mapping[int, T_Name],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        multihot: Tensor,
    ) -> List[List[T_Name]]:
        names = multihot_to_multinames(multihot, self.idx_to_name)
        return names


class MultinamesToIndices(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.multinames_to_indices`.
    """

    def __init__(
        self,
        idx_to_name: Mapping[int, T_Name],
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name

    def forward(
        self,
        names: List[List[T_Name]],
    ) -> List[List[int]]:
        indices = multinames_to_indices(names, self.idx_to_name)
        return indices


class MultinamesToMultihot(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.multinames_to_multihot`.
    """

    def __init__(
        self,
        idx_to_name: Mapping[int, T_Name],
        *,
        device: DeviceLike = None,
        dtype: DTypeLike = torch.bool,
    ) -> None:
        super().__init__()
        self.idx_to_name = idx_to_name
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        names: List[List[T_Name]],
    ) -> Tensor:
        multihot = multinames_to_multihot(
            names,
            self.idx_to_name,
            device=self.device,
            dtype=self.dtype,
        )
        return multihot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class ProbsToIndices(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.probs_to_indices`.
    """

    def __init__(
        self,
        threshold: Union[float, Tensor],
        *,
        padding_idx: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.threshold = threshold
        self.padding_idx = padding_idx

    def forward(
        self,
        probs: Tensor,
    ) -> List[List[int]]:
        indices = probs_to_indices(probs, self.threshold, padding_idx=self.padding_idx)
        return indices


class ProbsToMultihot(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.probs_to_multihot`.
    """

    def __init__(
        self,
        threshold: Union[float, Tensor],
        *,
        device: DeviceLike = None,
        dtype: DTypeLike = torch.bool,
    ) -> None:
        super().__init__()
        self.threshold = threshold
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        probs: Tensor,
    ) -> Tensor:
        multihot = probs_to_multihot(
            probs,
            self.threshold,
            device=self.device,
            dtype=self.dtype,
        )
        return multihot

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                device=self.device,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class ProbsToMultinames(Generic[T_Name], nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.multilabel.probs_to_multinames`.
    """

    def __init__(
        self,
        threshold: Union[float, Tensor],
        idx_to_name: Mapping[int, T_Name],
    ) -> None:
        super().__init__()
        self.threshold = threshold
        self.idx_to_name = idx_to_name

    def forward(
        self,
        probs: Tensor,
    ) -> List[List[T_Name]]:
        names = probs_to_multinames(probs, self.threshold, self.idx_to_name)
        return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchoutil.utils.pack.dataset import PackedDataset
from torchoutil.utils.pack.pack import (
    ExistsMode,
    pack_dataset,
    pack_dataset_dict,
    pack_dataset_to_columns,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

import torch
from torch import Generator, Tensor, nn
from torch.types import Number

from torchoutil.core.get import DeviceLike
from torchoutil.nn.functional.pad import (
    PadAlign,
    PadMode,
    PadValue,
    pad_and_stack_rec,
    pad_dim,
    pad_dims,
)
from torchoutil.pyoutil.collections import dump_dict


class PadDim(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.pad.pad_dim`.
    """

    def __init__(
        self,
        target_length: int,
        dim: int = -1,
        align: PadAlign = "left",
        pad_value: PadValue = 0.0,
        mode: PadMode = "constant",
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.target_length = target_length
        self.dim = dim
        self.align: PadAlign = align
        self.pad_value = pad_value
        self.mode: PadMode = mode
        self.generator = generator

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return pad_dim(
            x,
            target_length=self.target_length,
            dim=self.dim,
            align=self.align,
            pad_value=self.pad_value,
            mode=self.mode,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_length=self.target_length,
                dim=self.dim,
                align=self.align,
                pad_value=self.pad_value,
                mode=self.mode,
            )
        )


class PadDims(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.pad.pad_dims`.
    """

    def __init__(
        self,
        target_lengths: Iterable[int],
        dims: Iterable[int] = (-1,),
        aligns: Iterable[PadAlign] = ("left",),
        pad_value: PadValue = 0.0,
        mode: PadMode = "constant",
        generator: Union[int, Generator, None] = None,
    ) -> None:
        super().__init__()
        self.target_lengths = target_lengths
        self.aligns = aligns
        self.pad_value = pad_value
        self.dims = dims
        self.mode: PadMode = mode
        self.generator = generator

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return pad_dims(
            x,
            target_lengths=self.target_lengths,
            dims=self.dims,
            aligns=self.aligns,
            pad_value=self.pad_value,
            mode=self.mode,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_lengths=self.target_lengths,
                dims=self.dims,
                aligns=self.aligns,
                pad_value=self.pad_value,
                mode=self.mode,
            )
        )


class PadAndStackRec(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.pad.pad_and_stack_rec`.
    """

    def __init__(
        self,
        pad_value: Number = 0,
        *,
        align: PadAlign = "left",
        device: DeviceLike = None,
        dtype: Union[None, torch.dtype] = None,
    ) -> None:
        super().__init__()
        self.pad_value = pad_value
        self.align = align
        self.device = device
        self.dtype = dtype

    def forward(
        self,
        sequence: Union[Tensor, int, float, tuple, list],
    ) -> Tensor:
        return pad_and_stack_rec(
            sequence,
            self.pad_value,
            align=self.align,
            device=self.device,
            dtype=self.dtype,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                pad_value=self.pad_value,
                align=self.align,
                device=self.device,
                dtype=self.dtype,
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.modules.mixins import ESequential
from torchoutil.nn.modules.multilabel import (
    IndicesToMultihot,
    IndicesToMultinames,
    MultihotToIndices,
    MultihotToMultinames,
    MultinamesToIndices,
    MultinamesToMultihot,
)


class TestMultilabel(TestCase):
    def test_example_1(self) -> None:
        for _ in range(10):
            num_steps = int(torch.randint(1, 2, ()).item())
            num_samples = int(torch.randint(1, 2, ()).item())
            num_classes = int(torch.randint(1, 20, ()).item())
            threshold = torch.rand(())
            idx_to_name = dict(zip(range(num_classes), map(str, range(num_classes))))

            multihot = torch.rand(num_steps, num_samples, num_classes).ge(threshold)

            # dummy pipeline to convert labels multiple times
            pipeline = ESequential(
                MultihotToIndices(),
                IndicesToMultinames(idx_to_name),
                MultinamesToMultihot(idx_to_name),
                MultihotToMultinames(idx_to_name),
                MultinamesToIndices(idx_to_name),
                IndicesToMultihot(num_classes),
            )
            result = pipeline(multihot)

            assert torch.equal(multihot, result), f"{multihot.shape=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
import sys
from pathlib import Path
from re import Pattern
from typing import Any, Generator, Iterable, List, Tuple, Union

from torchoutil.pyoutil.logging import warn_once
from torchoutil.pyoutil.re import PatternLike, compile_patterns, match_patterns

pylog = logging.getLogger(__name__)


def get_num_cpus_available() -> int:
    """Returns the number of CPUs available for the current process on Linux-based platforms.

    On Windows and MAC OS, this will just return the number of logical CPUs on this machine.
    If the number of CPUs cannot be detected, returns 0.
    """
    try:
        num_cpus = len(os.sched_getaffinity(0))
    except AttributeError:
        msg = "Cannot detect number of CPUs available for the current process. This function will just returns the number of CPUs."
        warn_once(msg, __name__)

        num_cpus = os.cpu_count()
        if num_cpus is None:
            num_cpus = 0
    return num_cpus


def safe_rmdir(
    root: Union[str, Path],
    *,
    rm_root: bool = True,
    error_on_non_empty_dir: bool = True,
    followlinks: bool = False,
    dry_run: bool = False,
    verbose: int = 0,
) -> Tuple[List[str], List[str]]:
    """Remove all empty sub-directories.

    Args:
        root: Root directory path.
        rm_root: If True, remove the root directory too if it is empty at the end. defaults to True.
        error_on_non_empty_dir: If True, raises a RuntimeError if a subdirectory contains at least 1 file. Otherwise it will ignore non-empty directories. defaults to True.
        followlinks: Indicates whether or not symbolic links shound be followed. defaults to False.
        dry_run: If True, does not remove any directory and just output the list of directories which could be deleted. defaults to False.
        verbose: Verbose level. defaults to 0.

    Returns:
        A tuple containing the list of directories paths deleted and the list of directories paths reviewed.
    """
    root = str(root)
    if not osp.isdir(root):
        msg = f"Target root directory does not exists. (with {root=})"
        raise FileNotFoundError(msg)

    to_delete = {}
    reviewed = []
    walker = os.walk(root, topdown=False, followlinks=followlinks)

    for dpath, dnames, fnames in walker:
        reviewed.append(dpath)

        if not rm_root and dpath == root:
            continue

        elif len(fnames) == 0 and (
            all(osp.join(dpath, dname) in to_delete for dname in dnames)
        ):
            to_delete[dpath] = None

        elif error_on_non_empty_dir:
            raise RuntimeError(f"Cannot remove non-empty directory '{dpath}'.")
        elif verbose >= 2:
            pylog.debug(f"Ignoring non-empty directory '{dpath}'...")

    if not dry_run:
        for dpath in to_delete:
            os.rmdir(dpath)

    return list(to_delete), reviewed


def tree_iter(
    root: Union[str, Path],
    *,
    exclude: Union[PatternLike, Iterable[PatternLike]] = (),
    space: str = "    ",
    branch: str = "│   ",
    tee: str = "├── ",
    last: str = "└── ",
    max_depth: int = sys.maxsize,
    followlinks: bool = False,
) -> Generator[str, Any, None]:
    """A recursive generator, given a directory Path object will yield a visual tree structure line by line with each line prefixed by the same characters

    Based on: https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
    """
    root = Path(root)
    if not root.is_dir():
        raise ValueError(f"Invalid argument path '{root}'. (not a directory)")

    if not followlinks and root.is_symlink():
        yield from ()
        return

    exclude = compile_patterns(exclude)
    if match_patterns(str(root), exclude):
        yield from ()
        return

    yield root.resolve().name + "/"

    if max_depth <= 0:
        return

    yield from _tree_impl(
        root,
        exclude=exclude,
        prefix="",
        space=space,
        branch=branch,
        tee=tee,
        last=last,
        depth=1,
        max_depth=max_depth,
        followlinks=followlinks,
    )


def _tree_impl(
    root: Path,
    exclude: List[Pattern],
    prefix: str,
    space: str,
    branch: str,
    tee: str,
    last: str,
    depth: int,
    max_depth: int,
    followlinks: bool,
) -> Generator[str, Any, None]:
    paths = root.iterdir()
    try:
        paths = [
            path
            for path in paths
            if (followlinks or not path.is_symlink())
            and not match_patterns(str(path), exclude)
        ]
    except PermissionError:
        paths = []

    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(paths) - 1) + [last]

    for pointer, path in zip(pointers, paths):
        is_dir = path.is_dir()
        suffix = "/" if is_dir else ""
        yield prefix + pointer + path.name + suffix

        if is_dir and depth <= max_depth:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from _tree_impl(
                path,
                exclude=exclude,
                prefix=prefix + extension,
                space=space,
                branch=branch,
                tee=tee,
                last=last,
                depth=depth + 1,
                max_depth=max_depth,
                followlinks=followlinks,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchoutil.core.packaging import _NUMPY_AVAILABLE, _YAML_AVAILABLE

from .common import to_builtin
from .csv import load_csv, to_csv
from .json import load_json, to_json
from .pickle import load_pickle, to_pickle

if _NUMPY_AVAILABLE:
    from .save_fn import numpy_save_fn

if _YAML_AVAILABLE:
    from .yaml import load_yaml, to_yaml



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from torch import Tensor, nn

from torchoutil.nn.functional.mask import masked_mean, masked_sum
from torchoutil.pyoutil.collections import dump_dict


class MaskedMean(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.mask.masked_mean`.
    """

    def __init__(self, dim: Union[None, int, Iterable[int]] = None) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, tensor: Tensor, non_pad_mask: Tensor) -> Tensor:
        reduced = masked_mean(tensor, non_pad_mask, dim=self.dim)
        return reduced

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
            ),
            ignore_lst=(None,),
        )


class MaskedSum(nn.Module):
    """
    For more information, see :func:`~torchoutil.nn.functional.mask.masked_sum`.
    """

    def __init__(self, dim: Union[None, int, Iterable[int]] = None) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, tensor: Tensor, non_pad_mask: Tensor) -> Tensor:
        reduced = masked_sum(tensor, non_pad_mask, dim=self.dim)
        return reduced

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
            ),
            ignore_lst=(None,),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from typing import Callable, Iterable, List, Optional, Union

import torch
from torch import Generator, Tensor

from torchoutil.core.get import get_generator
from torchoutil.pyoutil.collections import flat_list_of_list


def random_split(
    num_samples: int,
    lengths: Iterable[float],
    generator: Union[int, Generator, None] = None,
) -> List[List[int]]:
    """Generate indices for a random dataset split.

    Args:
        num_samples: Number of total samples.
        lengths: Ratios of the target splits.
        generator: Torch Generator or seed to make this function deterministic. defaults to None.
    """
    lengths = _round_lengths(num_samples, lengths, math.floor)
    generator = get_generator(generator)

    indices = torch.randperm(num_samples, generator=generator)
    start = 0
    splits = []
    for length in lengths:
        end = start + length
        split = indices[start:end].tolist()
        splits.append(split)
    return splits


def balanced_monolabel_split(
    targets_indices: Tensor,
    num_classes: int,
    lengths: Iterable[float],
    generator: Union[int, Generator, None] = None,
) -> List[List[int]]:
    """Generate indices for a random dataset split while keeping the same multiclass distribution.

    Args:
        targets: List of class indices of size (N,).
        num_classes: Number of classes.
        lengths: Ratios of the target splits.
        generator: Torch Generator or seed to make this function deterministic. defaults to None.
    """
    lengths = list(lengths)
    generator = get_generator(generator)

    indices_per_class = []
    for class_idx in range(num_classes):
        indices = torch.where(targets_indices.eq(class_idx))[0]
        indices = indices.tolist()
        indices_per_class.append(indices)

    indices_per_class = _shuffle_indices_per_class(indices_per_class, generator)
    splits = _split_indices_per_class(indices_per_class, lengths, math.floor)
    flatten = [flat_list_of_list(split)[0] for split in splits]
    return flatten


def _round_lengths(
    n: int,
    lengths: Iterable[Union[int, float]],
    round_fn: Callable[[float], int],
) -> List[int]:
    int_lenghts = []
    for length in lengths:
        if isinstance(length, float):
            length = round_fn(n * length)
        int_lenghts.append(length)
    return int_lenghts


def _shuffle_indices_per_class(
    indices_per_class: List[List[int]],
    generator: Optional[Generator],
) -> List[List[int]]:
    shuffled = []
    for indices in indices_per_class:
        perm = torch.randperm(len(indices), generator=generator)
        indices = torch.as_tensor(indices)
        indices = indices[perm].tolist()
        shuffled.append(indices)
    return indices_per_class


def _split_indices_per_class(
    indices_per_class: List[List[int]],
    lengths: List[float],
    round_fn: Callable[[float], int],
) -> List[List[List[int]]]:
    """
    Split distinct indices per class.

    Example:
    --------
    ```
    >>> split_indices_per_class(indices_per_class=[[1, 2], [3, 4], [5, 6]], ratios=[0.5, 0.5])
    ... [[[1], [3], [5]], [[2], [4], [6]]]
    ```

    Args:
        indices_per_class: List of indices of each class.
        lengths: The ratios of each indices split.
    """
    assert 0.0 <= sum(lengths) <= 1.0, "Ratio sum cannot be greater than 1.0."

    num_classes = len(indices_per_class)
    num_splits = len(lengths)

    indices_per_ratio_per_class = [
        [[] for _ in range(num_classes)] for _ in range(num_splits)
    ]

    current_starts = [0 for _ in range(num_classes)]

    for i, length in enumerate(lengths):
        for j, indices in enumerate(indices_per_class):
            current_start = current_starts[j]
            current_end = current_start + round_fn(length * len(indices))
            sub_indices = indices[current_start:current_end]
            indices_per_ratio_per_class[i][j] = sub_indices
            current_starts[j] = current_end

    return indices_per_ratio_per_class



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import unittest
from typing import Union
from unittest import TestCase

from torchoutil.utils.data.slicer import DatasetSlicer, DatasetSlicerWrapper


class DummySequence1:
    def __len__(self) -> int:
        return 5

    def __getitem__(self, idx: int):
        return 2 * idx


class DummySequence2:
    def __len__(self) -> int:
        return 5

    def __getitem__(self, idx: Union[tuple, int]):
        if isinstance(idx, tuple):
            idx, *args = idx
        else:
            args = ()
        return (2 * idx,) + tuple(args)


class DummySequence3(DatasetSlicer):
    def __len__(self) -> int:
        return 5

    def get_item(self, idx: int) -> int:
        return idx * 3


class TestSlicer(TestCase):
    def test_slicer_wrapper(self) -> None:
        ds = DummySequence1()
        slicer = DatasetSlicerWrapper(ds)

        indices = list(range(len(ds)))
        mask = [True] * len(ds)
        items = [ds[i] for i in indices]

        assert items == slicer[:]
        assert items == slicer[indices]
        assert items == slicer[mask]

    def test_slicer_wrapper_args(self) -> None:
        ds = DummySequence2()
        slicer = DatasetSlicerWrapper(ds)

        indices = list(range(len(ds)))
        mask = [True] * len(ds)
        items = [ds[i] for i in indices]

        assert items == slicer[:]
        assert items == slicer[indices]
        assert items == slicer[mask]

        assert [(2, "a", "b"), (4, "a", "b")] == slicer[1:3, "a", "b"]

    def test_slicer_base_class(self) -> None:
        ds = DummySequence3()

        indices = list(range(len(ds)))
        random.shuffle(indices)

        mask = [random.random() > 0.5 for _ in range(len(ds))]

        assert ds[:] == [ds[i] for i in range(len(ds))]
        assert ds[indices] == [ds[idx] for idx in indices]
        assert ds[mask] == [ds[i] for i in range(len(ds)) if mask[i]]


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# MIT License
#
# Copyright (c) 2023- CNRS
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - https://herve.niderb.fr
# Alexis PLAQUET

# This code is based on PyAnnote source code from: https://github.com/pyannote/pyannote-audio/blob/4407a66023cb42fd74450ab83b802a09ffa27d52/pyannote/audio/utils/powerset.py

import math
from functools import lru_cache
from itertools import combinations
from typing import Optional, overload

import torch
from torch import Tensor
from torch.nn import functional as F

from torchoutil.core.get import DTypeLike, get_dtype
from torchoutil.nn.functional.multiclass import probs_to_onehot
from torchoutil.pyoutil.logging import warn_once
from torchoutil.types.tensor_subclasses import Tensor2D, Tensor3D


@overload
def multilabel_to_powerset(
    multilabel: Tensor,
    *,
    mapping: Tensor,
) -> Tensor3D:
    ...


@overload
def multilabel_to_powerset(
    multilabel: Tensor,
    *,
    num_classes: int,
    max_set_size: int,
) -> Tensor3D:
    ...


def multilabel_to_powerset(
    multilabel: Tensor,
    *,
    mapping: Optional[Tensor] = None,
    num_classes: Optional[int] = None,
    max_set_size: Optional[int] = None,
) -> Tensor3D:
    """
    Args:
        multilabel: (batch_size, num_frames, num_classes) Tensor

    Returns:
        powerset: (batch_size, num_frames, num_powerset_classes) Tensor
    """
    if mapping is not None:
        num_powerset_classes, _num_classes = mapping.shape
    elif num_classes is not None and max_set_size is not None:
        mapping = build_powerset_mapping(num_classes, max_set_size)
        num_powerset_classes, _ = mapping.shape
    else:
        msg = "Either mapping or (num_classes and max_set_size) must be provided as arguments, but all of them are None."
        raise ValueError(msg)

    if not multilabel.is_floating_point():
        tgt_dtype = mapping.dtype
        msg = f"Implicit multilabel conversion from {multilabel.dtype} to {tgt_dtype} in multilabel_to_powerset fn."
        warn_once(msg, __name__)
        multilabel = multilabel.to(dtype=tgt_dtype)

    powerset = F.one_hot(
        torch.argmax(torch.matmul(multilabel, mapping.T), dim=-1),
        num_classes=num_powerset_classes,
    )
    return powerset  # type: ignore


@overload
def powerset_to_multilabel(
    powerset: Tensor,
    soft: bool = False,
    *,
    mapping: Tensor,
) -> Tensor3D:
    ...


@overload
def powerset_to_multilabel(
    powerset: Tensor,
    soft: bool = False,
    *,
    num_classes: int,
    max_set_size: int,
) -> Tensor3D:
    ...


def powerset_to_multilabel(
    powerset: Tensor,
    soft: bool = False,
    *,
    mapping: Optional[Tensor] = None,
    num_classes: Optional[int] = None,
    max_set_size: Optional[int] = None,
) -> Tensor3D:
    """
    Args:
        powerset:  Powerset logits, probabilities or onehot tensor of shape (batch_size, num_frames, num_powerset_classes).

    Returns:
        multilabel: (batch_size, num_frames, num_classes) Tensor
    """
    if mapping is not None:
        pass
    elif num_classes is not None and max_set_size is not None:
        mapping = build_powerset_mapping(num_classes, max_set_size)
    else:
        msg = "Either mapping or (num_classes and max_set_size) must be provided as arguments, but all of them are None."
        raise ValueError(msg)

    if soft:
        powerset_probs = powerset.exp()
    else:
        powerset_probs = probs_to_onehot(powerset, dim=-1, dtype=torch.float)

    multilabel = torch.matmul(powerset_probs, mapping)
    return multilabel  # type: ignore


@lru_cache(maxsize=None)
def build_powerset_mapping(
    num_classes: int,
    max_set_size: int,
    dtype: DTypeLike = None,
) -> Tensor2D:
    """Build powerset mapping matrix of shape (num_powerset_classes, num_classes)."""
    dtype = get_dtype(dtype)
    num_powerset_classes = get_num_powerset_classes(num_classes, max_set_size)
    mapping = torch.zeros(num_powerset_classes, num_classes, dtype=dtype)
    powerset_k = 0

    for set_size in range(0, max_set_size + 1):
        for current_set in combinations(range(num_classes), set_size):
            mapping[powerset_k, current_set] = 1
            powerset_k += 1

    return mapping  # type: ignore


@lru_cache(maxsize=None)
def get_num_powerset_classes(num_classes: int, max_set_size: int) -> int:
    binom_coefficients = (math.comb(num_classes, i) for i in range(0, max_set_size + 1))
    return int(sum(binom_coefficients))
